# -*- coding: utf-8 -*-
"""
Implementation of VAE.

References
    - https://stackoverflow.com/questions/63987125/keras-lstm-vae-variational-autoencoder-for-time-series-anamoly-detection # noqa: E501
    - https://github.com/twairball/keras_lstm_vae/blob/master/lstm_vae/vae.py
"""
import logging
import tempfile

import numpy as np
import tensorflow as tf
from mlstars.utils import import_object
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model

LOGGER = logging.getLogger(__name__)


def build_layer(layer: dict, hyperparameters: dict):
    layer_class = import_object(layer['class'])
    layer_kwargs = layer['parameters'].copy()
    # TODO: Upgrade to using tf.keras.layers.Wrapper in mlprimitives.
    if issubclass(layer_class, tf.keras.layers.Wrapper):
        layer_kwargs['layer'] = build_layer(layer_kwargs['layer'], hyperparameters)
    for key, value in layer_kwargs.items():
        if isinstance(value, str):
            layer_kwargs[key] = hyperparameters.get(value, value)
    return layer_class(**layer_kwargs)


class VAE(object):
    """VAE model for time series reconstruction.

    Args:
        layers_encoder (list):
            List containing layers of encoder.
        layers_generator (list):
            List containing layers of generator.
        optimizer (str):
            String denoting the keras optimizer.
        input_shape (tuple):
            Optional. Tuple denoting the shape of an input sample.
        output_shape (tuple):
            Optional. Tuple denoting the shape of an output sample.
        learning_rate (float):
            Optional. Float denoting the learning rate of the optimizer. Default 0.005.
        epochs (int):
            Optional. Integer denoting the number of epochs. Default 2000.
        latent_dim (int):
            Optional. Integer denoting dimension of latent space. Default 20.
        batch_size (int):
            Integer denoting the batch size. Default 64.
        shuffle (bool):
            Optional. Whether to shuffle the training data before each epoch Default True.
        verbose (int):
            Verbosity mode where 0 = silent, 1 = progress bar, 2 = one line per epoch. Default 0.
        callbacks (list):
            Optional. List of keras callbacks to apply during evaluation.
        validation_split (float):
            Optional. Fraction of the training data to be used as validation data. Default 0.
        hyperparameters (dictionary):
            Optional. Dictionary containing any additional inputs.
    """

    def __getstate__(self):
        networks = ['encoder', 'generator']
        modules = ['optimizer', 'vae_model', 'mse_loss', 'fit_history']

        state = self.__dict__.copy()

        for module in modules:
            del state[module]

        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                tf.keras.models.save_model(state.pop(network), fd.name, overwrite=True)

                state[network + '_str'] = fd.read()

        return state

    def __setstate__(self, state):
        networks = ['encoder', 'generator']
        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                fd.write(state.pop(network + '_str'))
                fd.flush()

                state[network] = tf.keras.models.load_model(fd.name)

        self.__dict__ = state

    def _build_model(self, hyperparameters, layers, input_shape):
        x = Input(shape=input_shape)
        model = tf.keras.models.Sequential()

        for layer in layers:
            built_layer = build_layer(layer, hyperparameters)
            model.add(built_layer)

        return Model(x, model(x))

    def _setdefault(self, kwargs, key, value):
        if key in kwargs:
            return

        if key in self.hyperparameters and self.hyperparameters[key] is None:
            kwargs[key] = value

    def __init__(self, layers_encoder: list, layers_generator: list, optimizer: str,
                 input_shape: tuple = None, output_shape: tuple = None, latent_dim: int = 20,
                 learning_rate: float = 0.001, epochs: int = 35, batch_size: int = 64,
                 shuffle: bool = True, verbose: bool = True, callbacks: tuple = tuple(),
                 validation_split: float = 0.0, **hyperparameters):

        self.input_shape = input_shape
        self.latent_dim = latent_dim
        self.latent_shape = (latent_dim,)
        self.output_shape = output_shape

        self.layers_encoder = layers_encoder
        self.layers_generator = layers_generator

        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = import_object(optimizer)(learning_rate)
        self.mse_loss = tf.keras.losses.MeanSquaredError()
        self.shuffle = shuffle
        self.verbose = verbose
        self.hyperparameters = hyperparameters
        self.validation_split = validation_split
        for callback in callbacks:
            callback['class'] = import_object(callback['class'])
        self.callbacks = callbacks
        self._fitted = False
        self.fit_history = None

    def _augment_hyperparameters(self, X, y, kwargs):
        input_shape = np.asarray(X)[0].shape
        output_shape = np.asarray(y)[0].shape
        self.input_shape = self.input_shape or input_shape
        self.output_shape = self.output_shape or output_shape

        self._setdefault(kwargs, 'length', input_shape[0])
        self._setdefault(kwargs, 'output_dim', self.output_shape[-1])
        return kwargs

    def _set_shapes(self):
        self.encoder_input_shape = self.input_shape
        self.generator_input_shape = self.latent_shape

    def _sampling(self, args):
        z_mean, z_log_sigma = args
        batch_size = tf.shape(z_mean)[0]
        epsilon = K.random_normal(shape=(batch_size, self.latent_dim), mean=0., stddev=1.)
        return z_mean + z_log_sigma * epsilon

    def _vae_loss(self, y_true, y_pred, z_log_sigma, z_mean):
        rc_loss = self.mse_loss(y_true, y_pred)
        kl_loss = - 0.5 * K.mean(1 + z_log_sigma - K.square(z_mean) - K.exp(z_log_sigma))
        loss = rc_loss + kl_loss
        return loss

    def _build_vae(self, **kwargs):
        hyperparameters = self.hyperparameters.copy()
        hyperparameters.update(kwargs)

        self.encoder = self._build_model(
            hyperparameters, self.layers_encoder, self.encoder_input_shape)
        self.generator = self._build_model(
            hyperparameters, self.layers_generator, self.generator_input_shape)

        x = Input(shape=self.input_shape)
        y = Input(shape=self.output_shape)

        # Sample latent vector.
        h = self.encoder(x)
        z_mean = tf.keras.layers.Dense(self.latent_dim)(h)
        z_log_sigma = tf.keras.layers.Dense(self.latent_dim)(h)
        z = tf.keras.layers.Lambda(self._sampling)([z_mean, z_log_sigma])

        y_ = self.generator(z)

        self.vae_model = Model([x, y], y_)
        self.vae_model.add_loss(self._vae_loss(y, y_, z_log_sigma, z_mean))
        self.vae_model.compile(optimizer=self.optimizer)

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """Fit the model.

        Args:
            X (ndarray):
                N-dimensional array containing the input sequences for the model.
            y (ndarray):
                N-dimensional array containing the output sequences we want to reconstruct.
        """
        # Reconstruct the same input.
        if y is None:
            y = X.copy()

        if not self._fitted:
            self._augment_hyperparameters(X, y, kwargs)
            self._set_shapes()
            self._build_vae(**kwargs)

        callbacks = [
            callback['class'](**callback.get('args', dict()))
            for callback in self.callbacks
        ]

        self.fit_history = self.vae_model.fit((X, y),
                                              batch_size=self.batch_size,
                                              epochs=self.epochs,
                                              shuffle=self.shuffle,
                                              verbose=self.verbose,
                                              callbacks=callbacks,
                                              validation_split=self.validation_split,
                                              )
        self._fitted = True

    def predict(self, X: np.ndarray) -> tuple:
        """Predict values using the initialized object.

        Args:
            X (ndarray):
                N-dimensional array containing the input sequences for the model.

        Returns:
            ndarray:
                N-dimensional array containing the reconstructions for each input sequence.
        """
        return self.vae_model.predict(X)



# -*- coding: utf-8 -*-
import logging
import tempfile

import numpy as np
import tensorflow as tf
from mlstars.utils import import_object
from numpy import ndarray
from sklearn.preprocessing import MinMaxScaler
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model

from orion.primitives.timeseries_errors import reconstruction_errors, regression_errors

LOGGER = logging.getLogger(__name__)


def build_layer(layer: dict, hyperparameters: dict):
    layer_class = import_object(layer['class'])
    layer_kwargs = layer['parameters'].copy()
    # TODO: Upgrade to using tf.keras.layers.Wrapper in mlprimitives.
    if issubclass(layer_class, tf.keras.layers.Wrapper):
        layer_kwargs['layer'] = build_layer(layer_kwargs['layer'], hyperparameters)
    for key, value in layer_kwargs.items():
        if isinstance(value, str):
            layer_kwargs[key] = hyperparameters.get(value, value)
    return layer_class(**layer_kwargs)


class AER(object):
    """Autoencoder with bi-directional regression for time series anomaly detection.

    Args:
        layers_encoder (list):
            List containing layers of encoder.
        layers_generator (list):
            List containing layers of generator.
        optimizer (str):
            String denoting the keras optimizer.
        input_shape (tuple):
            Optional. Tuple denoting the shape of an input sample.
        learning_rate (float):
            Optional. Float denoting the learning rate of the optimizer. Default 0.005.
        epochs (int):
            Optional. Integer denoting the number of epochs. Default 2000.
        batch_size (int):
            Integer denoting the batch size. Default 64.
        hyperparameters (dictionary):
            Optional. Dictionary containing any additional inputs.
    """

    def __getstate__(self):
        networks = ['encoder', 'decoder']
        modules = ['aer_model', 'optimizer', 'fit_history']

        state = self.__dict__.copy()

        for module in modules:
            del state[module]

        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                tf.keras.models.save_model(state.pop(network), fd.name, overwrite=True)

                state[network + '_str'] = fd.read()

        return state

    def __setstate__(self, state):
        networks = ['encoder', 'decoder']
        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                fd.write(state.pop(network + '_str'))
                fd.flush()

                state[network] = tf.keras.models.load_model(fd.name)

        self.__dict__ = state

    @staticmethod
    def _build_model(hyperparameters, layers, input_shape):
        x = Input(shape=input_shape)
        model = tf.keras.models.Sequential()

        for layer in layers:
            built_layer = build_layer(layer, hyperparameters)
            model.add(built_layer)

        return Model(x, model(x))

    def __init__(self, layers_encoder: list, layers_decoder: list,
                 optimizer: str, learning_rate: float = 0.001,
                 epochs: int = 35, batch_size: int = 64, shuffle: bool = True,
                 verbose: bool = True, callbacks: tuple = tuple(), reg_ratio: float = 0.5,
                 lstm_units: int = 30, validation_split: float = 0.0, **hyperparameters):

        self.layers_encoder = layers_encoder
        self.layers_decoder = layers_decoder
        self.epochs = epochs
        self.batch_size = batch_size
        self.reg_ratio = reg_ratio
        self.optimizer = import_object(optimizer)(learning_rate)
        self.lstm_units = lstm_units

        self.shuffle = shuffle
        self.verbose = verbose
        self.hyperparameters = hyperparameters
        self.validation_split = validation_split
        for callback in callbacks:
            callback['class'] = import_object(callback['class'])
        self.callbacks = callbacks

        self._fitted = False
        self.fit_history = None

    def _augment_hyperparameters(self, X, y, kwargs):
        # Infers the shape from the inputs.
        self.input_shape = np.asarray(X)[0].shape
        self.latent_shape = (self.lstm_units * 2,)

        kwargs['repeat_vector_n'] = self.input_shape[0] + 2
        kwargs['lstm_units'] = self.lstm_units
        return kwargs

    def _build_aer(self, **kwargs):
        hyperparameters = self.hyperparameters.copy()
        hyperparameters.update(kwargs)

        # Build encoder and decoder.
        self.encoder = self._build_model(hyperparameters, self.layers_encoder,
                                         self.input_shape)
        self.decoder = self._build_model(hyperparameters, self.layers_decoder,
                                         self.latent_shape)

        # Build AER Model.
        x = tf.keras.Input(self.input_shape)
        x_ = self.decoder(self.encoder(x))
        ry, y, fy = x_[:, 0], x_[:, 1:-1], x_[:, -1]

        self.aer_model = tf.keras.Model(inputs=x, outputs=[ry, y, fy])
        self.aer_model.compile(
            loss=['mse', 'mse', 'mse'],
            optimizer=self.optimizer,
            loss_weights=[self.reg_ratio / 2, 1 - self.reg_ratio, self.reg_ratio / 2],
        )

    def fit(self, X: np.ndarray, y: np.ndarray = None, **kwargs):
        """Fit the model.

        Args:
            X (ndarray):
                N-dimensional array containing the input sequences for the model.
            y (ndarray):
                N-dimensional array containing the output sequences we want to reconstruct.
        """
        if y is None:
            y = X.copy()  # Reconstruct the same input.

        X = X[:, 1:-1, :]
        ry, y, fy = y[:, 0], y[:, 1:-1], y[:, -1]

        if not self._fitted:
            self._augment_hyperparameters(X, y, kwargs)
            self._build_aer(**kwargs)

        callbacks = [
            callback['class'](**callback.get('args', dict()))
            for callback in self.callbacks
        ]

        self.fit_history = self.aer_model.fit(
            X, [ry, y, fy], batch_size=self.batch_size, epochs=self.epochs,
            shuffle=self.shuffle, verbose=self.verbose, callbacks=callbacks,
            validation_split=self.validation_split
        )

        self._fitted = True

    def predict(self, X: np.ndarray) -> tuple:
        """Predict values using the initialized object.

        Args:
            X (ndarray):
                N-dimensional array containing the input sequences for the model.

        Returns:
            ndarray:
                N-dimensional array containing the regression for each input sequence (forward).
            ndarray:
                N-dimensional array containing the reconstructions for each input sequence.
            ndarray:
                N-dimensional array containing the regression for each input sequence (reverse).
        """
        X = X[:, 1:-1, :]
        x_ = self.decoder.predict(self.encoder.predict(X))
        ry, y, fy = x_[:, 0], x_[:, 1:-1], x_[:, -1]
        return ry, y, fy


def bi_regression_errors(y: ndarray, ry_hat: ndarray, fy_hat: ndarray,
                         smoothing_window: float = 0.01, smooth: bool = True, mask: bool = True):
    """Compute an array of absolute errors comparing the forward and reverse predictions with
    the expected output.

    Anomaly scores are created in the forward and reverse directions. Scores in overlapping indices
    are averaged while scores in non-overlapping indices are taken directly from either forward or
    reverse anomaly scores.

    If smooth is True, apply EWMA to the resulting array of errors.

    Args:
        y (ndarray):
            Ground truth.
        ry_hat (ndarray):
            Predicted values (reverse).
        fy_hat (ndarray):
            Predicted values (forward).
        smoothing_window (float):
            Optional. Size of the smoothing window, expressed as a proportion of the total
            length of y. If not given, 0.01 is used.
        smooth (bool):
            Optional. Indicates whether the returned errors should be smoothed with EWMA.
            If not given, `True` is used.
        mask (bool): bool = True
            Optional. Mask anomaly score errors in the beginning.
            If not given, `True` is used.

    Returns:
        ndarray:
            Array of errors.
    """
    time_steps = len(y[0]) - 1
    mask_steps = int(smoothing_window * len(y)) if mask else 0
    ry, fy = y[:, 0], y[:, -1]

    f_scores = regression_errors(fy, fy_hat, smoothing_window=smoothing_window, smooth=smooth)
    f_scores[:mask_steps] = 0
    f_scores = np.concatenate([np.zeros(time_steps), f_scores])

    r_scores = regression_errors(ry, ry_hat, smoothing_window=smoothing_window, smooth=smooth)
    r_scores[:mask_steps] = min(r_scores)
    r_scores = np.concatenate([r_scores, np.zeros(time_steps)])

    scores = f_scores + r_scores
    scores[time_steps + mask_steps:-time_steps] /= 2
    return scores


def score_anomalies(y: ndarray, ry_hat: ndarray, y_hat: ndarray, fy_hat: ndarray,
                    smoothing_window: float = 0.01, smooth: bool = True, mask: bool = True,
                    comb: str = 'mult', lambda_rec: float = 0.5, rec_error_type: str = "dtw"):
    """Compute an array of absolute errors comparing predictions and expected output.

    If smooth is True, apply EWMA to the resulting array of errors.

    Args:
        y (ndarray):
            Ground truth.
        ry_hat (ndarray):
            Predicted values (reverse).
        y_hat (ndarray):
            Predicted values. Each timestamp has multiple predictions.
        fy_hat (ndarray):
            Predicted values (forward).
        smoothing_window (float):
            Optional. Size of the smoothing window, expressed as a proportion of the total
            length of y. If not given, 0.01 is used.
        smooth (bool):
            Optional. Indicates whether the returned errors should be smoothed with EWMA.
            If not given, `True` is used.
        mask (bool): bool = True
            Optional. Mask anomaly score errors in the beginning.
            If not given, `True` is used.
        rec_error_type (str):
            Optional. The method to compute reconstruction error. Can be one of
            `["point", "area", "dtw"]`. If not given, 'dtw' is used.
        comb (str):
            Optional. How to combine critic and reconstruction error. Can be one
            of `["mult", "sum", "rec"]`. If not given, 'mult' is used.
        lambda_rec (float):
            Optional. Used if `comb="sum"` as a lambda weighted sum to combine
            scores. If not given, 0.5 is used.

    Returns:
        ndarray:
            Array of errors.
    """

    reg_scores = bi_regression_errors(y, ry_hat, fy_hat,
                                      smoothing_window=smoothing_window,
                                      smooth=smooth,
                                      mask=mask
                                      )
    rec_scores, _ = reconstruction_errors(y[:, 1:-1], y_hat,
                                          smoothing_window=smoothing_window,
                                          smooth=smooth,
                                          rec_error_type=rec_error_type)
    mask_steps = int(smoothing_window * len(y)) if mask else 0
    rec_scores[:mask_steps] = min(rec_scores)
    rec_scores = np.concatenate([np.zeros(1), rec_scores, np.zeros(1)])

    scores = None
    if comb == "mult":
        reg_scores = MinMaxScaler([1, 2]).fit_transform(reg_scores.reshape(-1, 1)).flatten()
        rec_scores = MinMaxScaler([1, 2]).fit_transform(rec_scores.reshape(-1, 1)).flatten()
        scores = np.multiply(reg_scores, rec_scores)

    elif comb == "sum":
        reg_scores = MinMaxScaler([0, 1]).fit_transform(reg_scores.reshape(-1, 1)).flatten()
        rec_scores = MinMaxScaler([0, 1]).fit_transform(rec_scores.reshape(-1, 1)).flatten()
        scores = (1 - lambda_rec) * reg_scores + lambda_rec * rec_scores

    elif comb == "rec":
        scores = rec_scores

    elif comb == "reg":
        scores = reg_scores

    return scores



"""
Time Series anomaly detection functions.
Some of the implementation is inspired by the paper https://arxiv.org/pdf/1802.04431.pdf
"""

import numpy as np
import pandas as pd
from scipy.optimize import fmin


def deltas(errors, epsilon, mean, std):
    """Compute mean and std deltas.

    delta_mean = mean(errors) - mean(all errors below epsilon)
    delta_std = std(errors) - std(all errors below epsilon)

    Args:
        errors (ndarray):
            Array of errors.
        epsilon (ndarray):
            Threshold value.
        mean (float):
            Mean of errors.
        std (float):
            Standard deviation of errors.

    Returns:
        float, float:
            * delta_mean.
            * delta_std.
    """
    below = errors[errors <= epsilon]
    if not len(below):
        return 0, 0

    return mean - below.mean(), std - below.std()


def count_above(errors, epsilon):
    """Count number of errors and continuous sequences above epsilon.

    Continuous sequences are counted by shifting and counting the number
    of positions where there was a change and the original value was true,
    which means that a sequence started at that position.

    Args:
        errors (ndarray):
            Array of errors.
        epsilon (ndarray):
            Threshold value.

    Returns:
        int, int:
            * Number of errors above epsilon.
            * Number of continuous sequences above epsilon.
    """
    above = errors > epsilon
    total_above = len(errors[above])

    above = pd.Series(above)
    shift = above.shift(1)
    change = above != shift

    total_consecutive = sum(above & change)

    return total_above, total_consecutive


def z_cost(z, errors, mean, std):
    """Compute how bad a z value is.

    The original formula is::

                 (delta_mean/mean) + (delta_std/std)
        ------------------------------------------------------
        number of errors above + (number of sequences above)^2

    which computes the "goodness" of `z`, meaning that the higher the value
    the better the `z`.

    In this case, we return this value inverted (we make it negative), to convert
    it into a cost function, as later on we will use scipy.fmin to minimize it.

    Args:
        z (ndarray):
            Value for which a cost score is calculated.
        errors (ndarray):
            Array of errors.
        mean (float):
            Mean of errors.
        std (float):
            Standard deviation of errors.

    Returns:
        float:
            Cost of z.
    """
    epsilon = mean + z * std

    delta_mean, delta_std = deltas(errors, epsilon, mean, std)
    above, consecutive = count_above(errors, epsilon)

    numerator = -(delta_mean / mean + delta_std / std)
    denominator = above + consecutive ** 2

    if denominator == 0:
        return np.inf

    return numerator / denominator


def _find_threshold(errors, z_range):
    """Find the ideal threshold.

    The ideal threshold is the one that minimizes the z_cost function. Scipy.fmin is used
    to find the minimum, using the values from z_range as starting points.

    Args:
        errors (ndarray):
            Array of errors.
        z_range (list):
            List of two values denoting the range out of which the start points for the
            scipy.fmin function are chosen.

    Returns:
        float:
            Calculated threshold value.
    """
    mean = errors.mean()
    std = errors.std()

    min_z, max_z = z_range
    best_z = min_z
    best_cost = np.inf
    for z in range(min_z, max_z):
        best = fmin(z_cost, z, args=(errors, mean, std), full_output=True, disp=False)
        z, cost = best[0:2]
        if cost < best_cost:
            best_z = z[0]
            best_cost = cost

    return mean + best_z * std


def _fixed_threshold(errors, k=4):
    """Calculate the threshold.

    The fixed threshold is defined as k standard deviations away from the mean.

    Args:
        errors (ndarray):
            Array of errors.

    Returns:
        float:
            Calculated threshold value.
    """
    mean = errors.mean()
    std = errors.std()

    return mean + k * std


def _find_sequences(errors, epsilon, anomaly_padding):
    """Find sequences of values that are above epsilon.

    This is done following this steps:

        * create a boolean mask that indicates which values are above epsilon.
        * mark certain range of errors around True values with a True as well.
        * shift this mask by one place, filing the empty gap with a False.
        * compare the shifted mask with the original one to see if there are changes.
        * Consider a sequence start any point which was true and has changed.
        * Consider a sequence end any point which was false and has changed.

    Args:
        errors (ndarray):
            Array of errors.
        epsilon (float):
            Threshold value. All errors above epsilon are considered an anomaly.
        anomaly_padding (int):
            Number of errors before and after a found anomaly that are added to the
            anomalous sequence.

    Returns:
        ndarray, float:
            * Array containing start, end of each found anomalous sequence.
            * Maximum error value that was not considered an anomaly.
    """
    above = pd.Series(errors > epsilon)
    index_above = np.argwhere(above.values)

    for idx in index_above.flatten():
        above[max(0, idx - anomaly_padding):min(idx + anomaly_padding + 1, len(above))] = True

    shift = above.shift(1).fillna(False)
    change = above != shift

    if above.all():
        max_below = 0
    else:
        max_below = max(errors[~above])

    index = above.index
    starts = index[above & change].tolist()
    ends = (index[~above & change] - 1).tolist()

    if len(ends) == len(starts) - 1:
        ends.append(len(above) - 1)

    return np.array([starts, ends]).T, max_below


def _get_max_errors(errors, sequences, max_below):
    """Get the maximum error for each anomalous sequence.

    Also add a row with the max error which was not considered anomalous.

    Table containing a ``max_error`` column with the maximum error of each
    sequence and the columns ``start`` and ``stop`` with the corresponding start and stop
    indexes, sorted descendingly by the maximum error.

    Args:
        errors (ndarray):
            Array of errors.
        sequences (ndarray):
            Array containing start, end of anomalous sequences
        max_below (float):
            Maximum error value that was not considered an anomaly.

    Returns:
        pandas.DataFrame:
            DataFrame object containing columns ``start``, ``stop`` and ``max_error``.
    """
    max_errors = [{
        'max_error': max_below,
        'start': -1,
        'stop': -1
    }]

    for sequence in sequences:
        start, stop = sequence
        sequence_errors = errors[start: stop + 1]
        max_errors.append({
            'start': start,
            'stop': stop,
            'max_error': max(sequence_errors)
        })

    max_errors = pd.DataFrame(max_errors).sort_values('max_error', ascending=False)
    return max_errors.reset_index(drop=True)


def _prune_anomalies(max_errors, min_percent):
    """Prune anomalies to mitigate false positives.

    This is done by following these steps:

        * Shift the errors 1 negative step to compare each value with the next one.
        * Drop the last row, which we do not want to compare.
        * Calculate the percentage increase for each row.
        * Find rows which are below ``min_percent``.
        * Find the index of the latest of such rows.
        * Get the values of all the sequences above that index.

    Args:
        max_errors (pandas.DataFrame):
            DataFrame object containing columns ``start``, ``stop`` and ``max_error``.
        min_percent (float):
            Percentage of separation the anomalies need to meet between themselves and the
            highest non-anomalous error in the window sequence.

    Returns:
        ndarray:
            Array containing start, end, max_error of the pruned anomalies.
    """
    next_error = max_errors['max_error'].shift(-1).iloc[:-1]
    max_error = max_errors['max_error'].iloc[:-1]

    increase = (max_error - next_error) / max_error
    too_small = increase < min_percent

    if too_small.all():
        last_index = -1
    else:
        last_index = max_error[~too_small].index[-1]

    return max_errors[['start', 'stop', 'max_error']].iloc[0: last_index + 1].values


def _compute_scores(pruned_anomalies, errors, threshold, window_start):
    """Compute the score of the anomalies.

    Calculate the score of the anomalies proportional to the maximum error in the sequence
    and add window_start timestamp to make the index absolute.

    Args:
        pruned_anomalies (ndarray):
            Array of anomalies containing the start, end and max_error for all anomalies in
            the window.
        errors (ndarray):
            Array of errors.
        threshold (float):
            Threshold value.
        window_start (int):
            Index of the first error value in the window.

    Returns:
        list:
            List of anomalies containing start-index, end-index, score for each anomaly.
    """
    anomalies = list()
    denominator = errors.mean() + errors.std()

    for row in pruned_anomalies:
        max_error = row[2]
        score = (max_error - threshold) / denominator
        anomalies.append([row[0] + window_start, row[1] + window_start, score])

    return anomalies


def _merge_sequences(sequences):
    """Merge consecutive and overlapping sequences.

    We iterate over a list of start, end, score triples and merge together
    overlapping or consecutive sequences.
    The score of a merged sequence is the average of the single scores,
    weighted by the length of the corresponding sequences.

    Args:
        sequences (list):
            List of anomalies, containing start-index, end-index, score for each anomaly.

    Returns:
        ndarray:
            Array containing start-index, end-index, score for each anomaly after merging.
    """
    if len(sequences) == 0:
        return np.array([])

    sorted_sequences = sorted(sequences, key=lambda entry: entry[0])
    new_sequences = [sorted_sequences[0]]
    score = [sorted_sequences[0][2]]
    weights = [sorted_sequences[0][1] - sorted_sequences[0][0]]

    for sequence in sorted_sequences[1:]:
        prev_sequence = new_sequences[-1]

        if sequence[0] <= prev_sequence[1] + 1:
            score.append(sequence[2])
            weights.append(sequence[1] - sequence[0])
            weighted_average = np.average(score, weights=weights)
            new_sequences[-1] = (prev_sequence[0], max(prev_sequence[1], sequence[1]),
                                 weighted_average)
        else:
            score = [sequence[2]]
            weights = [sequence[1] - sequence[0]]
            new_sequences.append(sequence)

    return np.array(new_sequences)


def _find_window_sequences(window, z_range, anomaly_padding, min_percent, window_start,
                           fixed_threshold):
    """Find sequences of values that are anomalous.

    We first find the threshold for the window, then find all sequences above that threshold.
    After that, we get the max errors of the sequences and prune the anomalies. Lastly, the
    score of the anomalies is computed.

    Args:
        window (ndarray):
            Array of errors in the window that is analyzed.
        z_range (list):
            List of two values denoting the range out of which the start points for the
            dynamic find_threshold function are chosen.
        anomaly_padding (int):
            Number of errors before and after a found anomaly that are added to the anomalous
            sequence.
        min_percent (float):
            Percentage of separation the anomalies need to meet between themselves and the
            highest non-anomalous error in the window sequence.
        window_start (int):
            Index of the first error value in the window.
        fixed_threshold (bool):
            Indicates whether to use fixed threshold or dynamic threshold.

    Returns:
        ndarray:
            Array containing the start-index, end-index, score for each anomalous sequence
            that was found in the window.
    """
    if fixed_threshold:
        threshold = _fixed_threshold(window)

    else:
        threshold = _find_threshold(window, z_range)

    window_sequences, max_below = _find_sequences(window, threshold, anomaly_padding)
    max_errors = _get_max_errors(window, window_sequences, max_below)
    pruned_anomalies = _prune_anomalies(max_errors, min_percent)
    window_sequences = _compute_scores(pruned_anomalies, window, threshold, window_start)

    return window_sequences


def find_anomalies(errors, index, z_range=(0, 10), window_size=None, window_size_portion=None,
                   window_step_size=None, window_step_size_portion=None, min_percent=0.1,
                   anomaly_padding=50, lower_threshold=False, fixed_threshold=None):
    """Find sequences of error values that are anomalous.

    We first define the window of errors, that we want to analyze. We then find the anomalous
    sequences in that window and store the start/stop index pairs that correspond to each
    sequence, along with its score. Optionally, we can flip the error sequence around the mean
    and apply the same procedure, allowing us to find unusually low error sequences.
    We then move the window and repeat the procedure.
    Lastly, we combine overlapping or consecutive sequences.

    Args:
        errors (ndarray):
            Array of errors.
        index (ndarray):
            Array of indices of the errors.
        z_range (list):
            Optional. List of two values denoting the range out of which the start points for
            the scipy.fmin function are chosen. If not given, (0, 10) is used.
        window_size (int):
            Optional. Size of the window for which a threshold is calculated. If not given,
            `None` is used, which finds one threshold for the entire sequence of errors.
        window_size_portion (float):
            Optional. Specify the size of the window to be a portion of the sequence of errors.
            If not given, `None` is used, and window size is used as is.
        window_step_size (int):
            Optional. Number of steps the window is moved before another threshold is
            calculated for the new window.
        window_step_size_portion (float):
            Optional. Specify the number of steps to be a portion of the window size. If not given,
            `None` is used, and window step size is used as is.
        min_percent (float):
            Optional. Percentage of separation the anomalies need to meet between themselves and
            the highest non-anomalous error in the window sequence. It nof given, 0.1 is used.
        anomaly_padding (int):
            Optional. Number of errors before and after a found anomaly that are added to the
            anomalous sequence. If not given, 50 is used.
        lower_threshold (bool):
            Optional. Indicates whether to apply a lower threshold to find unusually low errors.
            If not given, `False` is used.
        fixed_threshold (bool):
            Optional. Indicates whether to use fixed threshold or dynamic threshold. If not
            given, `False` is used.

    Returns:
        ndarray:
            Array containing start-index, end-index, score for each anomalous sequence that
            was found.
    """
    window_size = window_size or len(errors)
    if window_size_portion:
        window_size = np.ceil(len(errors) * window_size_portion).astype('int')

    window_step_size = window_step_size or window_size
    if window_step_size_portion:
        window_step_size = np.ceil(window_size * window_step_size_portion).astype('int')

    window_start = 0
    window_end = 0
    sequences = list()

    while window_end < len(errors):
        window_end = window_start + window_size
        window = errors[window_start:window_end]
        window_sequences = _find_window_sequences(window, z_range, anomaly_padding, min_percent,
                                                  window_start, fixed_threshold)
        sequences.extend(window_sequences)

        if lower_threshold:
            # Flip errors sequence around mean
            mean = window.mean()
            inverted_window = mean - (window - mean)
            inverted_window_sequences = _find_window_sequences(inverted_window, z_range,
                                                               anomaly_padding, min_percent,
                                                               window_start, fixed_threshold)
            sequences.extend(inverted_window_sequences)

        window_start = window_start + window_step_size

    sequences = _merge_sequences(sequences)

    anomalies = list()

    for start, stop, score in sequences:
        anomalies.append([index[int(start)], index[int(stop)], score])

    return np.asarray(anomalies)



from unittest import TestCase

import numpy as np
from numpy.testing import assert_allclose

from orion.primitives.timeseries_errors import (
    _area_error, _dtw_error, _point_wise_error, reconstruction_errors, regression_errors)


def test__point_wise_error():
    y = np.array([0.0, 0.1, 1.0, 0.5, 0.1, 0.1, 0.0, 0.5])
    y_hat = np.array([0.1, 2.0, 0.5, 0.0, 3.0, 0.1, 5.0, 0.5])

    expected = np.array([0.1, 1.9, 0.5, 0.5, 2.9, 0.0, 5.0, 0.0])
    returned = _point_wise_error(y, y_hat)

    np.testing.assert_array_equal(returned, expected)


def test__area_error():
    y = np.array([0.0, 0.1, 1.0, 0.5, 0.1, 0.1, 0.0, 0.5])
    y_hat = np.array([0.1, 2.0, 0.5, 0.0, 3.0, 0.1, 5.0, 0.5])
    score_window = 4

    expected = np.array([1.0, 1.7, 1.2, 1.4, 2.15, 5.15, 6.45, 5.0])
    returned = _area_error(y, y_hat, score_window)

    assert_allclose(returned, expected)


def test__dtw_error():
    y = np.array([0.0, 0.1, 1.0, 0.5])
    y_hat = np.array([0.1, 2.0, 0.5, 0.0])
    score_window = 2

    expected = np.array([0.0, 1.9, 0.0, 0.0])
    returned = _dtw_error(y, y_hat, score_window)

    assert_allclose(returned, expected)


class RegressionErrorsTest(TestCase):

    y = np.array([0.0, 0.1, 1.0, 0.5, 0.1, 0.1, 0.0, 0.5]).reshape(-1, 1)
    y_hat = np.array([0.1, 2.0, 0.5, 0.0, 3.0, 0.1, 5.0, 0.5]).reshape(-1, 1)

    def _run(self, smoothing_window, smooth, expected):
        sequences = regression_errors(self.y, self.y_hat, smoothing_window, smooth)

        assert_allclose(sequences, expected, rtol=1e-2)

    def test_no_smooth(self):
        smooth = False
        smoothing_window = 0
        expected = np.array([0.1, 1.9, 0.5, 0.5, 2.9, 0.0, 5.0, 0.0])
        self._run(smoothing_window, smooth, expected)

    def test_smooth(self):
        smooth = True
        smoothing_window = 0.125
        expected = np.array([0.1, 1.9, 0.5, 0.5, 2.9, 0.0, 5.0, 0.0])
        self._run(smoothing_window, smooth, expected)

    def test_smooth_span(self):
        smooth = True
        smoothing_window = 0.25
        expected = np.array([0.1, 1.45, 0.792, 0.595, 2.138, 0.71, 3.571, 1.19])
        self._run(smoothing_window, smooth, expected)


class ReconstructionErrorsTest(TestCase):

    y = np.array([
        [[0.0], [0.1]],
        [[1.0], [0.5]],
        [[0.1], [0.1]],
        [[0.0], [0.5]]
    ])

    y_hat = np.array([
        [[0.1], [2.0]],
        [[0.5], [0.0]],
        [[3.0], [0.1]],
        [[5.0], [0.5]]
    ])

    STEP_SIZE = 1

    def _run(self, score_window, smoothing_window, smooth, rec_error_type, expected):
        sequences, _ = reconstruction_errors(
            self.y, self.y_hat, self.STEP_SIZE, score_window, smoothing_window,
            smooth, rec_error_type)

        assert_allclose(sequences, expected, rtol=1e-2)

    def test_no_smooth(self):
        smooth = False
        score_window = 0
        smoothing_window = 0
        rec_error_type = 'point'
        expected = np.array([0.1, 0.25, 1.4, 2.55, 0.0])
        self._run(score_window, smoothing_window, smooth, rec_error_type, expected)

    def test_smooth(self):
        smooth = True
        score_window = 0
        smoothing_window = 0.25
        rec_error_type = 'point'
        expected = np.array([0.1, 0.25, 1.4, 2.55, 0.0])
        self._run(score_window, smoothing_window, smooth, rec_error_type, expected)

    def test_area(self):
        smooth = False
        score_window = 4
        smoothing_window = 0
        rec_error_type = 'area'
        expected = np.array([0.175, 1.0, 2.975, 4.075, 3.25])
        self._run(score_window, smoothing_window, smooth, rec_error_type, expected)

    def test_dtw(self):
        smooth = False
        score_window = 2
        smoothing_window = 0
        rec_error_type = 'dtw'
        expected = np.array([0.0, 0.27, 1.425, 0.0, 0.0])
        self._run(score_window, smoothing_window, smooth, rec_error_type, expected)



import json
import os
from unittest.mock import call, patch

import pandas as pd
import pytest
from mlblocks import MLPipeline
from mlblocks.discovery import load_pipeline

from orion import functional
from orion.core import Orion


class TestLoadData:

    @classmethod
    def setup_class(cls):
        cls.data = pd.DataFrame({
            'timestamp': list(range(100)),
            'value': [1] * 100,
        })

    def test_load_data_df(self):
        data = functional._load_data(self.data)

        assert data is self.data

    def test_load_data_path(self, tmpdir):
        path = os.path.join(tmpdir, 'data.csv')
        self.data.to_csv(path, index=False)

        data = functional._load_data(path)

        pd.testing.assert_frame_equal(data, self.data)

    def test_load_data_none(self):
        data = functional._load_data(None)

        assert data is None


class TestLoadDict:

    def test_load_dict_dict(self):
        a_dict = {'a': 'dict'}
        returned = functional._load_dict(a_dict)

        assert returned is a_dict

    def test_load_dict_json(self, tmpdir):
        a_dict = {'a': 'dict'}
        path = os.path.join(tmpdir, 'a.json')
        with open(path, 'w') as json_file:
            json.dump(a_dict, json_file)

        returned = functional._load_dict(path)

        assert returned == a_dict

    def test_load_dict_none(self):
        returned = functional._load_dict(None)

        assert returned is None


class TestLoadOrion:

    @classmethod
    def setup_class(cls):
        data = pd.DataFrame({
            'timestamp': list(range(100)),
            'value': [1] * 100,
        })
        cls.orion = Orion('dummy')
        cls.orion.fit(data)

    def test_load_orion_orion(self):
        orion = functional._load_orion(self.orion)

        assert orion is self.orion

    def test_load_orion_pickle(self, tmpdir):
        path = os.path.join(tmpdir, 'orion.pkl')
        self.orion.save(path)

        orion = functional._load_orion(path)

        assert orion is not self.orion
        assert orion == self.orion

    def test_load_orion_name(self):
        orion = functional._load_orion('dummy')

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert not orion._fitted
        assert orion._hyperparameters is None

    def test_load_orion_json_path(self, tmpdir):
        pipeline = load_pipeline('dummy')
        path = os.path.join(tmpdir, 'pipeline.json')
        with open(path, 'w') as json_file:
            json.dump(pipeline, json_file)

        orion = functional._load_orion(path)

        assert isinstance(orion, Orion)
        assert orion._pipeline == path
        assert not orion._fitted
        assert orion._hyperparameters is None

    def test_load_orion_dict(self):
        pipeline = load_pipeline('dummy')
        orion = functional._load_orion(pipeline)

        assert isinstance(orion, Orion)
        assert orion._pipeline == pipeline
        assert not orion._fitted
        assert orion._hyperparameters is None

    def test_load_orion_mlpipeline(self, tmpdir):
        pipeline = MLPipeline('dummy')

        orion = functional._load_orion(pipeline)

        assert isinstance(orion, Orion)
        assert orion._pipeline == pipeline
        assert not orion._fitted
        assert orion._hyperparameters is None

    def test_load_orion_hyperparams(self):
        hyperparams = {
            "orion.primitives.detectors.ThresholdDetector#1": {
                "ratio": 0.9
            }
        }
        orion = functional._load_orion('dummy', hyperparams)

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert not orion._fitted
        assert orion._hyperparameters == hyperparams

    def test_load_orion_invalid(self):
        with pytest.raises(ValueError):
            functional._load_orion('invalid')


class TestFitPipeline:

    @classmethod
    def setup_class(cls):
        cls.data = pd.DataFrame({
            'timestamp': list(range(100)),
            'value': [1] * 100,
        })

    @patch('orion.core.Orion.DEFAULT_PIPELINE', new='dummy')
    def test_fit_pipeline_default(self):
        orion = functional.fit_pipeline(self.data)

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert orion._fitted
        assert orion._hyperparameters is None

    def test_fit_pipeline_dict(self):
        pipeline = load_pipeline('dummy')

        orion = functional.fit_pipeline(self.data, pipeline)

        assert isinstance(orion, Orion)
        assert orion._pipeline == pipeline
        assert orion._fitted
        assert orion._hyperparameters is None

    def test_fit_pipeline_name(self):
        orion = functional.fit_pipeline(self.data, 'dummy')

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert orion._fitted
        assert orion._hyperparameters is None

    def test_fit_pipeline_csv(self, tmpdir):
        data_path = os.path.join(tmpdir, 'data.csv')
        self.data.to_csv(data_path, index=False)

        orion = functional.fit_pipeline(data_path, 'dummy')

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert orion._fitted
        assert orion._hyperparameters is None

    def test_fit_pipeline_hyperparams_dict(self):
        hyperparams = {
            "orion.primitives.detectors.ThresholdDetector#1": {
                "ratio": 0.9
            }
        }

        orion = functional.fit_pipeline(self.data, 'dummy', hyperparams)

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert orion._fitted
        assert orion._hyperparameters == hyperparams

    def test_fit_pipeline_hyperparams_json(self, tmpdir):
        hyperparams = {
            "orion.primitives.detectors.ThresholdDetector#1": {
                "ratio": 0.9
            }
        }
        hyperparams_path = os.path.join(tmpdir, 'hyperparams.json')
        with open(hyperparams_path, 'w') as json_file:
            json.dump(hyperparams, json_file)

        orion = functional.fit_pipeline(self.data, 'dummy', hyperparams_path)

        assert isinstance(orion, Orion)
        assert orion._pipeline == 'dummy'
        assert orion._fitted
        assert orion._hyperparameters == hyperparams

    def test_fit_pipeline_save_path(self, tmpdir):
        path = os.path.join(tmpdir, 'some/path.pkl')

        returned = functional.fit_pipeline(self.data, 'dummy', save_path=path)

        assert returned is None
        assert os.path.isfile(path)


class TestDetectAnomalies:

    @classmethod
    def setup_class(cls):
        cls.clean = pd.DataFrame({
            'timestamp': list(range(100)),
            'value': [1] * 100,
        })
        cls.anomalous = pd.DataFrame({
            'timestamp': list(range(100, 200)),
            'value': [1] * 45 + [10] * 10 + [1] * 45
        })
        cls.events = pd.DataFrame([
            {'start': 145, 'end': 155, 'severity': 9.0}
        ], columns=['start', 'end', 'severity'])

        cls.all_data = pd.concat((cls.clean, cls.anomalous))
        cls.all_events = pd.DataFrame([
            {'start': 145, 'end': 155, 'severity': 4.275}
        ], columns=['start', 'end', 'severity'])

    @patch('orion.core.Orion.DEFAULT_PIPELINE', new='dummy')
    def test_detect_anomalies_fit_default(self):
        anomalies = functional.detect_anomalies(
            data=self.anomalous,
            train_data=self.clean
        )

        pd.testing.assert_frame_equal(self.events, anomalies)

    def test_detect_anomalies_fit_pipeline(self):
        anomalies = functional.detect_anomalies(
            data=self.anomalous,
            pipeline='dummy',
            train_data=self.clean
        )

        pd.testing.assert_frame_equal(self.events, anomalies)

    @patch('orion.core.Orion.DEFAULT_PIPELINE', new='dummy')
    def test_detect_anomalies_fit_hyperparams(self):
        hyperparams = {
            "orion.primitives.detectors.ThresholdDetector#1": {
                "ratio": 0.9
            }
        }
        anomalies = functional.detect_anomalies(
            data=self.anomalous,
            hyperparameters=hyperparams,
            train_data=self.clean
        )

        pd.testing.assert_frame_equal(self.events, anomalies)

    def test_detect_anomalies_fit_pipeine_dict(self):
        pipeline = load_pipeline('dummy')
        anomalies = functional.detect_anomalies(
            data=self.anomalous,
            pipeline=pipeline,
            train_data=self.clean
        )

        pd.testing.assert_frame_equal(self.events, anomalies)

    def test_detect_anomalies_fitted_orion(self):
        orion = functional.fit_pipeline(self.clean, 'dummy')

        anomalies = functional.detect_anomalies(
            data=self.anomalous,
            pipeline=orion,
        )

        pd.testing.assert_frame_equal(self.events, anomalies)

    def test_detect_anomalies_saved_orion(self, tmpdir):
        orion_path = os.path.join(tmpdir, 'orion.pkl')
        functional.fit_pipeline(self.clean, 'dummy', save_path=orion_path)

        anomalies = functional.detect_anomalies(
            data=self.anomalous,
            pipeline=orion_path,
        )

        pd.testing.assert_frame_equal(self.events, anomalies)


class TestEvaluatePipeline:

    @patch('orion.functional._load_orion')
    @patch('orion.functional._load_data')
    def test_evaluate_pipeline_no_fit(self, load_data_mock, load_orion_mock):
        load_data_mock.side_effect = lambda x: x

        ret = functional.evaluate_pipeline('data', 'truth', 'pipeline', 'hyperparams', 'metrics')

        load_data_calls = [
            call('data'),
            call('truth'),
        ]
        assert load_data_calls == load_data_mock.call_args_list

        load_orion_mock.assert_called_once_with('pipeline', 'hyperparams')

        orion = load_orion_mock.return_value
        orion.detect.assert_called_once_with('data', 'truth', False, None, 'metrics')
        assert ret == orion.detect.return_value

    @patch('orion.functional._load_orion')
    @patch('orion.functional._load_data')
    def test_evaluate_pipeline_fit(self, load_data_mock, load_orion_mock):
        load_data_mock.side_effect = lambda x: x

        ret = functional.evaluate_pipeline(
            'data', 'truth', 'pipeline', 'hyperparams', 'metrics', 'train_data')

        load_data_calls = [
            call('data'),
            call('truth'),
            call('train_data'),
        ]
        assert load_data_calls == load_data_mock.call_args_list

        load_orion_mock.assert_called_once_with('pipeline', 'hyperparams')

        orion = load_orion_mock.return_value
        orion.detect.assert_called_once_with('data', 'truth', True, 'train_data', 'metrics')
        assert ret == orion.detect.return_value



"""Summarizing the results of the benchmark and producing leaderboard."""


import pandas as pd

DATASET_FAMILY = {
    "MSL": "NASA",
    "SMAP": "NASA",
    "YAHOOA1": "YAHOO",
    "YAHOOA2": "YAHOO",
    "YAHOOA3": "YAHOO",
    "YAHOOA4": "YAHOO",
    "artificialWithAnomaly": "NAB",
    "realAWSCloudwatch": "NAB",
    "realAdExchange": "NAB",
    "realTraffic": "NAB",
    "realTweets": "NAB"
}

DATASET_ABBREVIATION = {
    "MSL": "MSL",
    "SMAP": "SMAP",
    "YAHOOA1": "A1",
    "YAHOOA2": "A2",
    "YAHOOA3": "A3",
    "YAHOOA4": "A4",
    "artificialWithAnomaly": "Art",
    "realAWSCloudwatch": "AWS",
    "realAdExchange": "AdEx",
    "realTraffic": "Traf",
    "realTweets": "Tweets"
}


def get_f1_scores(results):
    df = results.groupby(['dataset', 'pipeline'])[['fp', 'fn', 'tp']].sum().reset_index()

    precision = df['tp'] / (df['tp'] + df['fp'])
    recall = df['tp'] / (df['tp'] + df['fn'])
    df['f1'] = 2 * (precision * recall) / (precision + recall)

    df = df.set_index(['dataset', 'pipeline'])[['f1']].unstack().T.droplevel(0)

    df.columns = [DATASET_ABBREVIATION[col] for col in df.columns]
    df.columns = pd.MultiIndex.from_tuples(list(zip(DATASET_FAMILY.values(), df.columns)))

    df['mean'] = df.mean(axis=1)
    df['std'] = df.std(axis=1)

    df.insert(0, 'Pipeline', df.index)
    df = df.reset_index(drop=True)

    return df


def get_summary_page(results):
    def get_status(x):
        return {
            "OK": 0,
            "ERROR": 1
        }[x]

    results['status'] = results['status'].apply(get_status)
    df = results.groupby(['dataset', 'pipeline'])[['fp', 'fn', 'tp']].sum().reset_index()

    precision = df['tp'] / (df['tp'] + df['fp'])
    recall = df['tp'] / (df['tp'] + df['fn'])
    df['f1'] = 2 * (precision * recall) / (precision + recall)

    summary = dict()

    # number of wins over arima
    arima_pipeline = 'arima'
    intermediate = df.set_index(['pipeline', 'dataset'])['f1'].unstack().T
    arima = intermediate.pop(arima_pipeline)

    summary['# Wins'] = (intermediate.T > arima).sum(axis=1)
    summary['# Wins'][arima_pipeline] = None

    # number of anomalies detected
    summary['# Anomalies'] = df.groupby('pipeline')[['tp', 'fp']].sum().sum(axis=1).to_dict()

    # average f1 score
    summary['Average F1 Score'] = df.groupby('pipeline')['f1'].mean().to_dict()

    # failure rate
    summary['Failure Rate'] = results.groupby(
        ['dataset', 'pipeline'])['status'].mean().unstack('pipeline').T.mean(axis=1)

    summary = pd.DataFrame(summary)
    summary.index.name = 'Pipeline'

    rank = 'Average F1 Score'
    summary.sort_values(rank, ascending=False, inplace=True)

    return summary.reset_index()


def add_sheet(df, name, writer, cell_fmt, header_fmt):
    widths = [0]
    startrow = 0
    offset = 1

    df_ = df.copy()
    if isinstance(df.columns, pd.MultiIndex):
        offset += 1
        df_.columns = df_.columns.droplevel()

    df_.to_excel(
        writer,
        sheet_name=name,
        startrow=startrow +
        offset,
        index=False,
        header=False,
        float_format="%0.4f")

    worksheet = writer.sheets[name]

    for idx, columns in enumerate(df.columns):
        column_name = columns
        if not isinstance(columns, tuple):
            columns = (columns, )

        for offset, column in enumerate(columns):
            worksheet.write(startrow + offset, idx, column, header_fmt)

            width = max(len(column), *df[column_name].astype(str).str.len()) + 1
            if len(widths) > idx:
                widths[idx] = max(widths[idx], width)
            else:
                widths.append(width)

    if isinstance(df.columns, pd.MultiIndex):
        columns = df.columns

        # horizontal
        worksheet.merge_range(0, 1, 0, 2, columns[1][0], header_fmt)
        worksheet.merge_range(0, 3, 0, 6, columns[3][0], header_fmt)
        worksheet.merge_range(0, 7, 0, 11, columns[7][0], header_fmt)

        # vertical
        worksheet.merge_range(0, 0, 1, 0, columns[0][0], header_fmt)
        worksheet.merge_range(0, 12, 1, 12, columns[12][0], header_fmt)
        worksheet.merge_range(0, 13, 1, 13, columns[13][0], header_fmt)

    for idx, width in enumerate(widths):
        worksheet.set_column(idx, idx, width + 1, cell_fmt)


def write_results(results, output, version):
    writer = pd.ExcelWriter(output, engine='xlsxwriter')
    cell_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10"
    })
    header_fmt = writer.book.add_format({
        "font_name": "Arial",
        "font_size": "10",
        "bold": True,
        "bottom": 1,
        "align": "center"
    })

    summary_page = get_summary_page(results)
    add_sheet(summary_page, version + '-Overview', writer, cell_fmt, header_fmt)
    f1_scores_page = get_f1_scores(results)
    add_sheet(f1_scores_page, version + '-F1-Scores', writer, cell_fmt, header_fmt)

    writer.save()



"""
Time Series error calculation functions.
"""

import math

import numpy as np
import pandas as pd
from pyts.metrics import dtw
from scipy import integrate


def regression_errors(y, y_hat, smoothing_window=0.01, smooth=True,
                      masking_window=0.01, mask=False):
    """Compute an array of absolute errors comparing predictions and expected output.

    If smooth is True, apply EWMA to the resulting array of errors.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values.
        smoothing_window (float):
            Optional. Size of the smoothing window, expressed as a proportion of the total
            length of y. If not given, 0.01 is used.
        smooth (bool):
            Optional. Indicates whether the returned errors should be smoothed with EWMA.
            If not given, `True` is used.
        masking_window (float):
            Optional. Size of the masking window, expressed as a proportion of the total
            length of y. If not given, 0.01 is used.
        mask (bool):
            Optional. Mask the start of anomaly scores.
            If not given, `False` is used.

    Returns:
        ndarray:
            Array of errors.
    """
    errors = np.abs(y - y_hat)[:, 0]

    if not smooth:
        return errors

    smoothing_window = max(1, int(len(y) * smoothing_window))
    errors = pd.Series(errors).ewm(span=smoothing_window).mean().values

    if mask:
        mask_length = int(masking_window * len(errors))
        errors[:mask_length] = min(errors)
    return errors


def _point_wise_error(y, y_hat):
    """Compute point-wise error between predicted and expected values.

    The computed error is calculated as the difference between predicted
    and expected values with a rolling smoothing factor.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values.

    Returns:
        ndarray:
            An array of smoothed point-wise error.
    """
    return abs(y - y_hat)


def _area_error(y, y_hat, score_window=10):
    """Compute area error between predicted and expected values.

    The computed error is calculated as the area difference between predicted
    and expected values with a smoothing factor.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values.
        score_window (int):
            Optional. Size of the window over which the scores are calculated.
            If not given, 10 is used.

    Returns:
        ndarray:
            An array of area error.
    """
    smooth_y = pd.Series(y).rolling(
        score_window, center=True, min_periods=score_window // 2).apply(integrate.trapz)
    smooth_y_hat = pd.Series(y_hat).rolling(
        score_window, center=True, min_periods=score_window // 2).apply(integrate.trapz)

    errors = abs(smooth_y - smooth_y_hat)

    return errors


def _dtw_error(y, y_hat, score_window=10):
    """Compute dtw error between predicted and expected values.

    The computed error is calculated as the dynamic time warping distance
    between predicted and expected values with a smoothing factor.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values.
        score_window (int):
            Optional. Size of the window over which the scores are calculated.
            If not given, 10 is used.

    Returns:
        ndarray:
            An array of dtw error.
    """
    length_dtw = (score_window // 2) * 2 + 1
    half_length_dtw = length_dtw // 2

    # add padding
    y_pad = np.pad(y, (half_length_dtw, half_length_dtw),
                   'constant', constant_values=(0, 0))
    y_hat_pad = np.pad(y_hat, (half_length_dtw, half_length_dtw),
                       'constant', constant_values=(0, 0))

    i = 0
    similarity_dtw = list()
    while i < len(y) - length_dtw:
        true_data = y_pad[i:i + length_dtw]
        true_data = true_data.flatten()

        pred_data = y_hat_pad[i:i + length_dtw]
        pred_data = pred_data.flatten()

        dist = dtw(true_data, pred_data)
        similarity_dtw.append(dist)
        i += 1

    errors = ([0] * half_length_dtw + similarity_dtw +
              [0] * (len(y) - len(similarity_dtw) - half_length_dtw))

    return errors


def reconstruction_errors(y, y_hat, step_size=1, score_window=10, smoothing_window=0.01,
                          smooth=True, rec_error_type='point'):
    """Compute an array of reconstruction errors.

    Compute the discrepancies between the expected and the
    predicted values according to the reconstruction error type.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values. Each timestamp has multiple predictions.
        step_size (int):
            Optional. Indicating the number of steps between windows in the predicted values.
            If not given, 1 is used.
        score_window (int):
            Optional. Size of the window over which the scores are calculated.
            If not given, 10 is used.
        smoothing_window (float or int):
            Optional. Size of the smoothing window, when float it is expressed as a proportion
            of the total length of y. If not given, 0.01 is used.
        smooth (bool):
            Optional. Indicates whether the returned errors should be smoothed.
            If not given, `True` is used.
        rec_error_type (str):
            Optional. Reconstruction error types ``["point", "area", "dtw"]``.
            If not given, "point" is used.

    Returns:
        ndarray:
            Array of reconstruction errors.
    """
    if isinstance(smoothing_window, float):
        smoothing_window = min(math.trunc(len(y) * smoothing_window), 200)

    true = [item[0] for item in y.reshape((y.shape[0], -1))]
    for item in y[-1][1:]:
        true.extend(item)

    predictions = []
    predictions_vs = []

    pred_length = y_hat.shape[1]
    num_errors = y_hat.shape[1] + step_size * (y_hat.shape[0] - 1)

    for i in range(num_errors):
        intermediate = []
        for j in range(max(0, i - num_errors + pred_length), min(i + 1, pred_length)):
            intermediate.append(y_hat[i - j, j])
        if intermediate:
            predictions.append(np.median(np.asarray(intermediate)))

            predictions_vs.append([[
                np.min(np.asarray(intermediate)),
                np.percentile(np.asarray(intermediate), 25),
                np.percentile(np.asarray(intermediate), 50),
                np.percentile(np.asarray(intermediate), 75),
                np.max(np.asarray(intermediate))
            ]])

    true = np.asarray(true)
    predictions = np.asarray(predictions)
    predictions_vs = np.asarray(predictions_vs)

    # Compute reconstruction errors
    if rec_error_type.lower() == "point":
        errors = _point_wise_error(true, predictions)

    elif rec_error_type.lower() == "area":
        errors = _area_error(true, predictions, score_window)

    elif rec_error_type.lower() == "dtw":
        errors = _dtw_error(true, predictions, score_window)

    # Apply smoothing
    if smooth:
        errors = pd.Series(errors).rolling(
            smoothing_window, center=True, min_periods=smoothing_window // 2).mean().values

    return errors, predictions_vs



import numpy as np
import pandas as pd


def slice_array_by_dims(X, target_index, axis=0):
    """Slice an ndarray by given dimensions.

    This function creates a copy of X then slices the array from the specificed
    dimension and returns the sliced array of the specified index.

    Args:
        X (ndarray):
            Array of input sequence.
        target_index (int or list[int]):
            Integer of the index to extract. Can be a list of integer values
            to extract multiple dimensions.
        axis (int or str):
            Optional. Axis along which to extract value. Default is ``0``.

    Returns:
        ndarray:
            Array of sliced values.
    """
    if isinstance(target_index, int):
        target_index = [target_index]

    dims = len(X.shape)
    if axis >= (dims):
        raise ValueError("Axis {} is outside the dimensions of X ({}).".format(axis, dims))

    indices = [slice(None)] * dims
    indices[axis] = target_index

    return X[tuple(indices)].copy()


def fillna(X, value=None, method=None, axis=None, limit=None, downcast=None):
    """Impute missing values.

    This function fills the missing values of the input sequence with the next/
    previous known value. If there are contigous NaN values, they will all be
    filled with the same next/previous known value.

    Args:
        X (ndarray or pandas.DataFrame):
            Array of input sequence.
        value:
            Optional. Value to use to fill holes (e.g. 0), alternately a dict/Series/DataFrame of
            values specifying which value to use for each index (for a Series) or column
            (for a DataFrame). Values not in the dict/Series/DataFrame will not be filled.
            This value cannot be a list. Default is None.
        method (str or list):
            Optional. String or list of strings describing whether to use forward or backward
            fill. pad / ffill: propagate last valid observation forward to next valid.
            backfill / bfill: use next valid observation to fill gap. Otherwise use ``None`` to
            fill with desired value. Possible values include
            ``[‘backfill’, ‘bfill’, ‘pad’, ‘ffill’, None]``. Default is None.
        axis (int or str):
            Optional. Axis along which to fill missing value. Possible values include 0 or
            "index", 1 or "columns". Default is None.
        limit (int):
            Optional. If method is specified, this is the maximum number of consecutive NaN values
            to forward/backward fill. In other words, if there is a gap with more than this number
            of consecutive NaNs, it will only be partially filled. If method is not specified,
            this is the maximum number of entries along the entire axis where NaNs will be filled.
            Must be greater than 0 if not None. Default is None.
        downcast (dict):
            Optional. A dict of item->dtype of what to downcast if possible, or the string "infer"
            which will try to downcast to an appropriate equal type (e.g. float64 to int64 if
            possible). Default is None.

    Returns:
        ndarray:
            Array of input sequence with imputed values.
    """
    if isinstance(method, str) or method is None:
        method = [method]

    if isinstance(X, np.ndarray):
        if X.ndim == 1:
            X_ = pd.Series(X)
        else:
            X_ = pd.DataFrame(X)

    else:
        X_ = X.copy()

    for fill in method:
        X_ = X_.fillna(value=value, method=fill, axis=axis, limit=limit, downcast=downcast)

    return X_.values



# -*- coding: utf-8 -*-

import ast
import concurrent
import json
import logging
import os
import pickle
import uuid
import warnings
from copy import deepcopy
from datetime import datetime
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
import tqdm
from scipy import signal as scipy_signal

from orion.analysis import _load_pipeline, analyze
from orion.data import load_anomalies, load_signal
from orion.evaluation import CONTEXTUAL_METRICS as METRICS
from orion.evaluation import contextual_confusion_matrix
from orion.progress import TqdmLogger, progress

warnings.simplefilter('ignore')

LOGGER = logging.getLogger(__name__)

BUCKET = 'sintel-orion'
S3_URL = 'https://{}.s3.amazonaws.com/{}'

BENCHMARK_PATH = os.path.join(os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '..'),
    'benchmark'
)

BENCHMARK_DATA = pd.read_csv(S3_URL.format(
    BUCKET, 'datasets.csv'), index_col=0, header=None).applymap(ast.literal_eval).to_dict()[1]
BENCHMARK_PARAMS = pd.read_csv(S3_URL.format(
    BUCKET, 'parameters.csv'), index_col=0, header=None).applymap(ast.literal_eval).to_dict()[1]

PIPELINE_DIR = os.path.join(os.path.dirname(__file__), 'pipelines', 'verified')

VERIFIED_PIPELINES = [
    'arima', 'lstm_dynamic_threshold', 'azure', 'tadgan',
    'aer', 'lstm_autoencoder', 'dense_autoencoder', 'vae'
]

VERIFIED_PIPELINES_GPU = {
    'arima': 'arima',
    'lstm_dynamic_threshold': 'lstm_dynamic_threshold',
    'azure': 'azure',
    'tadgan': 'tadgan_without_dropout_gpu',
    'aer': 'aer',
    'lstm_autoencoder': 'lstm_autoencoder',
    'dense_autoencoder': 'dense_autoencoder',
    'vae': 'vae'
}


def _load_signal(signal, test_split):
    if isinstance(test_split, float):
        train, test = load_signal(signal, test_size=test_split)
    elif test_split:
        train = load_signal(signal + '-train')
        test = load_signal(signal + '-test')
    else:
        train = test = load_signal(signal)

    return train, test


def _detrend_signal(df, value_column):
    df[value_column] = scipy_signal.detrend(df[value_column])
    return df


def _get_pipeline_hyperparameter(hyperparameters, dataset_name, pipeline_name):
    hyperparameters_ = deepcopy(hyperparameters)

    if hyperparameters:
        hyperparameters_ = hyperparameters_.get(dataset_name) or hyperparameters_
        hyperparameters_ = hyperparameters_.get(pipeline_name) or hyperparameters_

    if hyperparameters_ is None and dataset_name and pipeline_name:
        file_path = os.path.join(
            PIPELINE_DIR, pipeline_name, pipeline_name + '_' + dataset_name.lower() + '.json')
        if os.path.exists(file_path):
            hyperparameters_ = file_path

    if isinstance(hyperparameters_, str) and os.path.exists(hyperparameters_):
        with open(hyperparameters_) as f:
            hyperparameters_ = json.load(f)

    return hyperparameters_


def _parse_confusion_matrix(scores, truth):
    columns = ["tn", "fp", "fn", "tp"]
    metric_ = 'confusion_matrix'

    values = scores[metric_]
    if values == 0:
        fn = len(truth)
        values = (None, 0, fn, 0)  # (tn, fp, fn, tp)

    # formating output
    for metric_name, score in zip(columns, list(values)):
        scores[metric_name] = score

    del scores[metric_]


def _sort_leaderboard(df, rank, metrics):
    if rank not in df.columns:
        rank_ = list(metrics.keys())[0]
        LOGGER.exception("Rank %s is not in %s, using %s instead.",
                         rank, df.columns, rank_)
        rank = rank_

    df.sort_values(rank, ascending=False, inplace=True)
    df.reset_index(drop=True, inplace=True)
    df.index.name = 'rank'
    df.reset_index(drop=False, inplace=True)
    df['rank'] += 1

    pipeline_col = df.pop('pipeline')
    df.insert(0, 'pipeline', pipeline_col)
    return df


def _evaluate_signal(pipeline, signal, hyperparameter, metrics,
                     test_split=False, detrend=False, pipeline_path=None):

    train, test = _load_signal(signal, test_split)
    truth = load_anomalies(signal)

    if detrend:
        train = _detrend_signal(train, 'value')
        test = _detrend_signal(test, 'value')

    try:
        LOGGER.info("Scoring pipeline %s on signal %s (test split: %s)",
                    pipeline, signal, test_split)

        start = datetime.utcnow()
        pipeline = _load_pipeline(pipeline, hyperparameter)
        anomalies = analyze(pipeline, train, test)
        elapsed = datetime.utcnow() - start

        scores = {
            name: scorer(truth, anomalies, test)
            for name, scorer in metrics.items()
        }

        status = 'OK'

    except Exception as ex:
        LOGGER.exception("Exception scoring pipeline %s on signal %s (test split: %s), error %s.",
                         pipeline, signal, test_split, ex)

        elapsed = datetime.utcnow() - start
        scores = {
            name: 0 for name in metrics.keys()
        }

        status = 'ERROR'

    if 'confusion_matrix' in metrics.keys():
        _parse_confusion_matrix(scores, truth)

    scores['status'] = status
    scores['elapsed'] = elapsed.total_seconds()
    scores['split'] = test_split

    if pipeline_path:
        with open(pipeline_path, 'wb') as f:
            pickle.dump(pipeline, f)

    return scores


def _run_job(args):
    # Reset random seed
    np.random.seed()

    (pipeline, pipeline_name, dataset, signal, hyperparameter, metrics, test_split, detrend,
        iteration, cache_dir, pipeline_dir, run_id) = args

    pipeline_path = pipeline_dir
    if pipeline_dir:
        base_path = str(pipeline_dir / f'{pipeline_name}_{signal}_{dataset}_{iteration}')
        pipeline_path = base_path + '_pipeline.pkl'

    LOGGER.info('Evaluating pipeline %s on signal %s dataset %s (test split: %s); iteration %s',
                pipeline_name, signal, dataset, test_split, iteration)

    output = _evaluate_signal(
        pipeline,
        signal,
        hyperparameter,
        metrics,
        test_split,
        detrend,
        pipeline_path
    )
    scores = pd.DataFrame.from_records([output], columns=output.keys())

    scores.insert(0, 'dataset', dataset)
    scores.insert(1, 'pipeline', pipeline_name)
    scores.insert(2, 'signal', signal)
    scores.insert(3, 'iteration', iteration)
    scores['run_id'] = run_id

    if cache_dir:
        base_path = str(cache_dir / f'{pipeline_name}_{signal}_{dataset}_{iteration}_{run_id}')
        scores.to_csv(base_path + '_scores.csv', index=False)

    return scores


def _run_on_dask(jobs, verbose):
    """Run the tasks in parallel using dask."""
    try:
        import dask
    except ImportError as ie:
        ie.msg += (
            '\n\nIt seems like `dask` is not installed.\n'
            'Please install `dask` and `distributed` using:\n'
            '\n    pip install dask distributed'
        )
        raise

    scorer = dask.delayed(_run_job)
    persisted = dask.persist(*[scorer(args) for args in jobs])
    if verbose:
        try:
            progress(persisted)
        except ValueError:
            pass

    return dask.compute(*persisted)


def benchmark(pipelines=None, datasets=None, hyperparameters=None, metrics=METRICS, rank='f1',
              test_split=False, detrend=False, iterations=1, workers=1, show_progress=False,
              cache_dir=None, output_path=None, pipeline_dir=None):
    """Run pipelines on the given datasets and evaluate the performance.

    The pipelines are used to analyze the given signals and later on the
    detected anomalies are scored against the known anomalies using the
    indicated metrics.

    Finally, the scores obtained with each metric are averaged accross all the signals,
    ranked by the indicated metric and returned on a ``pandas.DataFrame``.

    Args:
        pipelines (dict or list): dictionary with pipeline names as keys and their
            JSON paths as values. If a list is given, it should be of JSON paths,
            and the paths themselves will be used as names. If not give, all verified
            pipelines will be used for evaluation.
        datasets (dict or list): dictionary of dataset name as keys and list of signals as
            values. If a list is given then it will be under a generic name ``dataset``.
            If not given, all benchmark datasets will be used used.
        hyperparameters (dict or list): dictionary with pipeline names as keys
            and their hyperparameter JSON paths or dictionaries as values. If a list is
            given, it should be of corresponding order to pipelines.
        metrics (dict or list): dictionary with metric names as keys and
            scoring functions as values. If a list is given, it should be of scoring
            functions, and they ``__name__`` value will be used as the metric name.
            If not given, all the available metrics will be used.
        rank (str): Sort and rank the pipelines based on the given metric.
            If not given, rank using the first metric.
        test_split (bool or float): Whether to use the prespecified train-test split. If
            float, then it should be between 0.0 and 1.0 and represent the proportion of
            the signal to include in the test split. If not given, use ``False``.
        detrend (bool): Whether to use ``scipy.detrend``. If not given, use ``False``.
        iterations (int):
            Number of iterations to perform over each signal and pipeline. Defaults to 1.
        workers (int or str):
            If ``workers`` is given as an integer value other than 0 or 1, a multiprocessing
            Pool is used to distribute the computation across the indicated number of workers.
            If the string ``dask`` is given, the computation is distributed using ``dask``.
            In this case, setting up the ``dask`` cluster and client is expected to be handled
            outside of this function.
        show_progress (bool):
            Whether to use tqdm to keep track of the progress. Defaults to ``True``.
        cache_dir (str):
            If a ``cache_dir`` is given, intermediate results are stored in the indicated directory
            as CSV files as they get computted. This allows inspecting results while the benchmark
            is still running and also recovering results in case the process does not finish
            properly. Defaults to ``None``.
        output_path (str): Location to save the intermediatry results. If not given,
            intermediatry results will not be saved.
        pipeline_dir (str):
            If a ``pipeline_dir`` is given, pipelines will get dumped in the specificed directory
            as pickle files. Defaults to ``None``.

    Returns:
        pandas.DataFrame:
            A table containing the scores obtained with each scoring function accross
            all the signals for each pipeline.
    """
    pipelines = pipelines or VERIFIED_PIPELINES
    datasets = datasets or BENCHMARK_DATA
    run_id = os.getenv('RUN_ID') or str(uuid.uuid4())[:10]

    if isinstance(pipelines, list):
        pipelines = {pipeline: pipeline for pipeline in pipelines}

    if isinstance(datasets, list):
        datasets = {'dataset': datasets}

    if isinstance(hyperparameters, list):
        hyperparameters = {pipeline: hyperparameter for pipeline, hyperparameter in
                           zip(pipelines.keys(), hyperparameters)}

    if isinstance(metrics, list):
        metrics_ = dict()
        for metric in metrics:
            if callable(metric):
                metrics_[metric.__name__] = metric
            elif metric in METRICS:
                metrics_[metric] = METRICS[metric]
            else:
                raise ValueError('Unknown metric: {}'.format(metric))

        metrics = metrics_

    if cache_dir:
        cache_dir = Path(cache_dir)
        os.makedirs(cache_dir, exist_ok=True)

    if pipeline_dir:
        pipeline_dir = Path(pipeline_dir)
        os.makedirs(pipeline_dir, exist_ok=True)

    jobs = list()
    for dataset, signals in datasets.items():
        for pipeline_name, pipeline in pipelines.items():
            hyperparameter = _get_pipeline_hyperparameter(hyperparameters, dataset, pipeline_name)
            parameters = BENCHMARK_PARAMS.get(dataset)
            if parameters is not None:
                detrend, test_split = parameters.values()
            for signal in signals:
                for iteration in range(iterations):
                    args = (
                        pipeline,
                        pipeline_name,
                        dataset,
                        signal,
                        hyperparameter,
                        metrics,
                        test_split,
                        detrend,
                        iteration,
                        cache_dir,
                        pipeline_dir,
                        run_id,
                    )
                    jobs.append(args)

    if workers == 'dask':
        scores = _run_on_dask(jobs, show_progress)
    else:
        if workers in (0, 1):
            scores = map(_run_job, jobs)
        else:
            pool = concurrent.futures.ProcessPoolExecutor(workers)
            scores = pool.map(_run_job, jobs)

        scores = tqdm.tqdm(scores, total=len(jobs), file=TqdmLogger())
        if show_progress:
            scores = tqdm.tqdm(scores, total=len(jobs))

    scores = pd.concat(scores)
    if output_path:
        LOGGER.info('Saving benchmark report to %s', output_path)
        scores.to_csv(output_path, index=False)

    return _sort_leaderboard(scores, rank, metrics)


def main(workers=1):
    pipeline_dir = 'save_pipelines'
    cache_dir = 'cache'

    # output path
    version = "results.csv"
    output_path = os.path.join(BENCHMARK_PATH, 'results', version)

    # metrics
    del METRICS['accuracy']
    METRICS['confusion_matrix'] = contextual_confusion_matrix
    metrics = {k: partial(fun, weighted=False) for k, fun in METRICS.items()}

    # pipelines
    pipelines = VERIFIED_PIPELINES

    results = benchmark(
        pipelines=pipelines, metrics=metrics, output_path=output_path, workers=workers,
        show_progress=False, pipeline_dir=pipeline_dir, cache_dir=cache_dir)

    return results


if __name__ == "__main__":
    results = main()



import pandas as pd
import pytest

from orion.evaluation.utils import (
    from_list_points_labels, from_list_points_timestamps, from_pandas_contextual,
    from_pandas_points, from_pandas_points_labels)


def assert_list_tuples(returned, expected_return):
    assert len(returned) == len(expected_return)
    for ret, exp_ret in zip(returned, expected_return):
        assert tuple(ret) == exp_ret


def test_from_pandas_contextual():
    anomalies = pd.DataFrame({'start': [2, 8], 'end': [5, 9]})

    expected_return = [(2, 5), (8, 9)]
    returned = from_pandas_contextual(anomalies)
    assert_list_tuples(returned, expected_return)


def test_from_pandas_contextual_severity():
    anomalies = pd.DataFrame({'start': [2, 8], 'end': [5, 9],
                              'severity': [0.1, 0.2]})

    expected_return = [(2, 5, 0.1), (8, 9, 0.2)]
    returned = from_pandas_contextual(anomalies)
    assert_list_tuples(returned, expected_return)


def test_from_pandas_contextual_error():
    anomalies = pd.DataFrame({'start': [2, 8]})
    with pytest.raises(KeyError):
        from_pandas_contextual(anomalies)


def test_from_list_points_timestamps():
    anomalies = [2, 3, 4, 5, 8, 9]

    expected_return = [(2, 5), (8, 9)]
    returned = from_list_points_timestamps(anomalies)
    assert_list_tuples(returned, expected_return)


def test_from_pandas_points():
    anomalies = pd.DataFrame({'timestamp': [2, 3, 4, 5, 8, 9]})

    expected_return = [(2, 5), (8, 9)]
    returned = from_pandas_points(anomalies)
    assert_list_tuples(returned, expected_return)


def test_from_pandas_points_error():
    anomalies = pd.DataFrame({'label': [0, 1]})
    with pytest.raises(KeyError):
        from_pandas_points(anomalies)


def test_from_pandas_points_labels():
    anomalies = pd.DataFrame({'timestamp': [2, 3, 4, 5, 6, 7, 8, 9],
                              'label': [1, 1, 1, 1, 0, 0, 1, 1]})

    expected_return = [(2, 5), (8, 9)]
    returned = from_pandas_points_labels(anomalies)
    assert_list_tuples(returned, expected_return)


def test_from_pandas_points_labels_error():
    anomalies = pd.DataFrame({'timestamp': [2, 8]})
    with pytest.raises(KeyError):
        from_pandas_points_labels(anomalies)


def test_from_list_points_labels():
    anomalies = [0, 0, 1, 1, 1, 1, 0, 0, 1, 1]

    expected_return = [(2, 5), (8, 9)]
    returned = from_list_points_labels(anomalies)
    assert_list_tuples(returned, expected_return)



"""Orion Functional API.

This module provides a collection of simple python functions that
allow using Orion performing as little steps as possible, hidding
away all the complexity related to loading data, creating class
instances or serializing and de-serializing previously fitted
pipelines.

Currently implemented functions:
- `fit_pipeline`: Learn an Orion pipeline and save it for later usage.
- `detect_anomalies`: Analyze a signal to detect anomalies. Optionally learn
  a pipeline on the way.
- `evaluate_pipeline`: Evaluate the performance of a pipeline against a list
  of known anomalies.
"""

import json
import os
from pickle import UnpicklingError
from typing import List, Union

import pandas as pd
from mlblocks import MLPipeline

from orion.core import Orion


def _load_data(data):
    if isinstance(data, pd.DataFrame):
        return data
    elif isinstance(data, str):
        return pd.read_csv(data)


def _load_dict(path_or_dict):
    if isinstance(path_or_dict, dict):
        return path_or_dict
    elif isinstance(path_or_dict, str) and os.path.exists(path_or_dict):
        with open(path_or_dict) as json_file:
            return json.load(json_file)


def _load_orion(pipeline, hyperparameters=None):
    if pipeline is None:
        return Orion()
    elif isinstance(pipeline, Orion):
        return pipeline
    else:
        hyperparameters = _load_dict(hyperparameters)
        try:
            return Orion(pipeline, hyperparameters)
        except ValueError:
            try:
                return Orion.load(pipeline)
            except (FileNotFoundError, UnpicklingError):
                raise ValueError('Invalid pipeline: {}'.format(pipeline))


def fit_pipeline(data: Union[str, pd.DataFrame],
                 pipeline: Union[str, MLPipeline, dict] = None,
                 hyperparameters: Union[str, pd.DataFrame] = None,
                 save_path: str = None) -> Orion:
    """Fit an Orion pipeline to the data.

    The pipeine can be passed as:
        * An ``str`` with a path to a JSON file.
        * An ``str`` with the name of a registered Orion pipeline.
        * An ``MLPipeline`` instance.
        * A ``dict`` with an ``MLPipeline`` specification.

    If no pipeline is passed, the default Orion pipeline is used.

    Args:
        data (str or DataFrame):
            Data to which the pipeline should be fitted.
            It can be passed as a path to a CSV file or as a DataFrame.
        pipeline (str, Pipeline or dict):
            Pipeline to use. It can be passed as:
                * An ``str`` with a path to a JSON file.
                * An ``str`` with the name of a registered pipeline.
                * An ``MLPipeline`` instance.
                * A ``dict`` with an ``MLPipeline`` specification.
        hyperparameters (str or dict):
            Hyperparameters to set to the pipeline. It can be passed as a
            hyperparameters ``dict`` in the ``mlblocks`` format or as a
            path to the corresponding JSON file. Defaults to
            ``None``.
        save_path (str):
            Path to the file where the fitted Orion instance will be stored
            using ``pickle``. If not given, the Orion instance is returned.
            Defaults to ``None``.

    Returns:
        Orion:
            If no save_path is provided, the fitted Orion instance is returned.
    """
    data = _load_data(data)
    hyperparameters = _load_dict(hyperparameters)

    if pipeline is None:
        pipeline = Orion.DEFAULT_PIPELINE

    orion = Orion(pipeline, hyperparameters)

    orion.fit(data)

    if save_path:
        orion.save(save_path)
    else:
        return orion


def detect_anomalies(data: Union[str, pd.DataFrame] = None,
                     pipeline: Union[Orion, str, MLPipeline, dict] = None,
                     hyperparameters: Union[str, pd.DataFrame] = None,
                     train_data: Union[str, pd.DataFrame] = None) -> pd.DataFrame:
    """Detect anomalies on timeseries data.

    The anomalies are detected using an Orion pipeline which can
    be passed as:

        * An ``Orion`` instance.
        * An ``str`` with the path to an Orion pickle file.
        * An ``str`` with a path to a JSON file.
        * An ``str`` with the name of a registered Orion pipeline.
        * An ``MLPipeline`` instance.
        * A ``dict`` with an ``MLPipeline`` specification.

    If no pipeline is passed, the default Orion pipeline is used.

    Optionally, separated learning data can be passed to fit
    the pipeline to it before using it to detect anomalies.

    Args:
        data (str or DataFrame):
            Data to analyze searching for anomalies.
            It can be passed as a path to a CSV file or as a DataFrame.
        pipeline (str or Pipeline or dict):
            Pipeline to use. It can be passed as:
                * An ``Orion`` instance.
                * An ``str`` with the path to an Orion pickle file.
                * An ``str`` with a path to a JSON file.
                * An ``str`` with the name of a registered pipeline.
                * An ``MLPipeline`` instance.
                * A ``dict`` with an ``MLPipeline`` specification.
        hyperparameters (str or dict):
            Hyperparameters to set to the pipeline. It can be passed as a
            hyperparameters ``dict`` in the ``mlblocks`` format or as a
            path to the corresponding JSON file. Ignored if being passed a
            previously serialized ``Orion`` instance. Defaults to ``None``.
        train_data (str or DataFrame):
            Data to which the pipeline should be fitted.
            It can be passed as a path to a CSV file or as a DataFrame.
            If not given, the pipeline is used without fitting it first.

    Returns:
        DataFrame:
            ``pandas.DataFrame`` containing the detected anomalies.
    """
    data = _load_data(data)
    orion = _load_orion(pipeline, hyperparameters)

    if train_data is not None:
        train_data = _load_data(train_data)
        orion.fit(train_data)

    return orion.detect(data)


def evaluate_pipeline(data: Union[str, pd.DataFrame],
                      truth: Union[str, pd.DataFrame],
                      pipeline: Union[str, dict, MLPipeline],
                      hyperparameters: Union[str, pd.DataFrame] = None,
                      metrics: List[Union[callable, str]] = None,
                      train_data: Union[str, pd.DataFrame] = None) -> pd.DataFrame:
    """Evaluate the performance of a pipeline.

    The pipeline is evaluated by executing it on a signal
    for which anomalies are known and then applying one or
    more metrics to it to compute scores.

    The pipeline can be passed as:
        * An ``str`` with a path to a JSON file.
        * An ``str`` with the path to a pickle file.
        * An ``str`` with the name of a registered Orion pipeline.
        * An ``MLPipeline`` instance.
        * A ``dict`` with an ``MLPipeline`` specification.

    If the pipeline is not fitted, it is possible to pass separated
    learning data can be passed to fit the pipeline to it before using
    it to detect anomalies.

    Args:
        data (str or DataFrame):
            Data to analyze searching for anomalies.
            It can be passed as a path to a CSV file or as a DataFrame.
        truth (str or DataFrame):
            Table of known anomalies to use as the ground truth for
            scoring. It can be passed as a path to a CSV file or as a
            DataFrame.
        pipeline (str or Pipeline or dict):
            Pipeline to use. It can be passed as:
                * An ``str`` with a path to a JSON file.
                * An ``str`` with the name of a registered pipeline.
                * An ``str`` with the path to a pickle file.
                * An ``MLPipeline`` instance.
                * A ``dict`` with an ``MLPipeline`` specification.
        hyperparameters (str or dict):
            Hyperparameters to set to the pipeline. It can be passed as
            a hyperparameters ``dict`` in the ``mlblocks`` format or as
            a path to the corresponding JSON file. Defaults to ``None``.
        metrics (list[str]):
            List of metrics to use. If not passed, all the Orion metrics
            are applied.
        train_data (str or DataFrame):
            Data to which the pipeline should be fitted.
            It can be passed as a path to a CSV file or as a DataFrame.
            If not given, the pipeline is used without fitting it first.
    """
    data = _load_data(data)
    truth = _load_data(truth)
    fit = train_data is not None
    if fit:
        train_data = _load_data(train_data)

    orion = _load_orion(pipeline, hyperparameters)

    return orion.detect(data, truth, fit, train_data, metrics)



"""Orion Core module.

This module defines the Orion Class, which is responsible for the
main anomaly detection functionality, as well as the interaction
with the underlying MLBlocks pipelines.
"""
import json
import logging
import os
import pickle
from typing import List, Union

import pandas as pd
from mlblocks import MLPipeline

from orion.evaluation import CONTEXTUAL_METRICS as METRICS

LOGGER = logging.getLogger(__name__)


class Orion:
    """Orion Class.

    The Orion Class provides the main anomaly detection functionalities
    of Orion and is responsible for the interaction with the underlying
    MLBlocks pipelines.

    Args:
        pipeline (str, dict or MLPipeline):
            Pipeline to use. It can be passed as:
                * An ``str`` with a path to a JSON file.
                * An ``str`` with the name of a registered pipeline.
                * An ``MLPipeline`` instance.
                * A ``dict`` with an ``MLPipeline`` specification.
        hyperparameters (dict):
            Additional hyperparameters to set to the Pipeline.
    """

    PIPELINES_DIR = tuple(
        dirname
        for dirname, _, _ in os.walk(os.path.join(os.path.dirname(__file__), 'pipelines'))
        if os.path.exists(os.path.join(dirname, os.path.basename(dirname) + '.json'))
    )
    PIPELINES = tuple(os.path.basename(pipeline) for pipeline in PIPELINES_DIR)

    DEFAULT_PIPELINE = 'lstm_dynamic_threshold'

    def _get_mlpipeline(self):
        pipeline = self._pipeline
        if isinstance(pipeline, str) and os.path.isfile(pipeline):
            with open(pipeline) as json_file:
                pipeline = json.load(json_file)

        mlpipeline = MLPipeline(pipeline)
        if self._hyperparameters:
            mlpipeline.set_hyperparameters(self._hyperparameters)

        return mlpipeline

    def __init__(self, pipeline: Union[str, dict, MLPipeline] = None,
                 hyperparameters: dict = None):
        self._pipeline = pipeline or self.DEFAULT_PIPELINE
        self._hyperparameters = hyperparameters
        self._mlpipeline = self._get_mlpipeline()
        self._fitted = False

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self._pipeline == other._pipeline and
            self._hyperparameters == other._hyperparameters and
            self._fitted == other._fitted
        )

    def __repr__(self):
        if isinstance(self._pipeline, MLPipeline):
            pipeline = '\n'.join(
                '    {}'.format(primitive) for primitive in self._pipeline.to_dict()['primitives'])

        elif isinstance(self._pipeline, dict):
            pipeline = '\n'.join(
                '    {}'.format(primitive) for primitive in self._pipeline['primitives'])

        else:
            pipeline = '    {}'.format(self._pipeline)

        hyperparameters = None
        if self._hyperparameters is not None:
            hyperparameters = '\n'.join(
                '    {}: {}'.format(step, value) for step, value in self._hyperparameters.items())

        return (
            'Orion:\n{}\n'
            'hyperparameters:\n{}\n'
        ).format(
            pipeline,
            hyperparameters
        )

    def fit(self, data: pd.DataFrame, **kwargs):
        """Fit the pipeline to the given data.

        Args:
            data (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                exactly two columns: timestamp and value.
        """
        if not self._fitted:
            self._mlpipeline = self._get_mlpipeline()

        self._mlpipeline.fit(data, **kwargs)
        self._fitted = True

    def _get_outputs_spec(self):
        outputs_spec = ["default"]
        try:
            visualization_outputs = self._mlpipeline.get_output_names('visualization')
            outputs_spec.append('visualization')
        except ValueError:
            visualization_outputs = []

        return outputs_spec, visualization_outputs

    @staticmethod
    def _build_events_df(events):
        events = pd.DataFrame(list(events), columns=['start', 'end', 'severity'])
        events['start'] = events['start'].astype('int64')
        events['end'] = events['end'].astype('int64')

        return events

    def _detect(self, method, data, visualization=False, **kwargs):
        if visualization:
            outputs_spec, visualization_names = self._get_outputs_spec()
        else:
            outputs_spec = 'default'

        outputs = method(data, output_=outputs_spec, **kwargs)

        if visualization:
            if visualization_names:
                events = outputs[0]
                visualization_outputs = outputs[-len(visualization_names):]
                visualization_dict = dict(zip(visualization_names, visualization_outputs))
            else:
                events = outputs
                visualization_dict = {}

            return self._build_events_df(events), visualization_dict

        return self._build_events_df(outputs)

    def detect(self, data: pd.DataFrame, visualization: bool = False) -> pd.DataFrame:
        """Detect anomalies in the given data..

        If ``visualization=True``, also return the visualization
        outputs from the MLPipeline object.

        Args:
            data (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                exactly two columns: timestamp and value.
            visualization (bool):
                If ``True``, also capture the ``visualization`` named
                output from the ``MLPipeline`` and return it as a second
                output.

        Returns:
            DataFrame or tuple:
                If visualization is ``False``, it returns the events
                DataFrame. If visualization is ``True``, it returns a
                tuple containing the events DataFrame followed by the
                visualization outputs dict.
        """
        return self._detect(self._mlpipeline.predict, data, visualization)

    def fit_detect(self, data: pd.DataFrame, visualization: bool = False,
                   **kwargs) -> pd.DataFrame:
        """Fit the pipeline to the data and then detect anomalies.

        This method is functionally equivalent to calling ``fit(data)``
        and later on ``detect(data)`` but with the difference that
        here the ``MLPipeline`` is called only once, using its ``fit``
        method, and the output is directly captured without having
        to execute the whole pipeline again during the ``predict`` phase.

        If ``visualization=True``, also return the visualization
        outputs from the MLPipeline object.

        Args:
            data (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                exactly two columns: timestamp and value.
            visualization (bool):
                If ``True``, also capture the ``visualization`` named
                output from the ``MLPipeline`` and return it as a second
                output.

        Returns:
            DataFrame or tuple:
                If visualization is ``False``, it returns the events
                DataFrame. If visualization is ``True``, it returns a
                tuple containing the events DataFrame followed by the
                visualization outputs dict.
        """
        if not self._fitted:
            self._mlpipeline = self._get_mlpipeline()

        result = self._detect(self._mlpipeline.fit, data, visualization, **kwargs)
        self._fitted = True

        return result

    def save(self, path: str):
        """Save this object using pickle.

        Args:
            path (str):
                Path to the file where the serialization of
                this object will be stored.
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as pickle_file:
            pickle.dump(self, pickle_file)

    @classmethod
    def load(cls, path: str):
        """Load an Orion instance from a pickle file.

        Args:
            path (str):
                Path to the file where the instance has been
                previously serialized.

        Returns:
            Orion

        Raises:
            ValueError:
                If the serialized object is not an Orion instance.
        """
        with open(path, 'rb') as pickle_file:
            orion = pickle.load(pickle_file)
            if not isinstance(orion, cls):
                raise ValueError('Serialized object is not an Orion instance')

            return orion

    def evaluate(self, data: pd.DataFrame, ground_truth: pd.DataFrame, fit: bool = False,
                 train_data: pd.DataFrame = None, metrics: List[str] = METRICS) -> pd.Series:
        """Evaluate the performance against ground truth anomalies.

        Args:
            data (DataFrame):
                Input data, passed as a ``pandas.DataFrame`` containing
                exactly two columns: timestamp and value.
            ground_truth (DataFrame):
                Ground truth anomalies passed as a ``pandas.DataFrame``
                containing two columns: start and stop.
            fit (bool):
                Whether to fit the pipeline before evaluating it.
                Defaults to ``False``.
            train_data (DataFrame):
                Training data, passed as a ``pandas.DataFrame`` containing
                exactly two columns: timestamp and value.
                If not given, the pipeline is fitted on ``data``.
            metrics (list):
                List of metrics to used passed as a list of strings.
                If not given, it defaults to all the Orion metrics.

        Returns:
            Series:
                ``pandas.Series`` containing one element for each
                metric applied, with the metric name as index.
        """
        if not fit:
            method = self._mlpipeline.predict
        else:
            if not self._fitted:
                mlpipeline = self._get_mlpipeline()
            else:
                mlpipeline = self._mlpipeline

            if train_data is not None:
                # Fit first and then predict
                mlpipeline.fit(train_data)
                method = mlpipeline.predict
            else:
                # Fit and predict at once
                method = mlpipeline.fit

        events = self._detect(method, data)

        scores = {
            metric: METRICS[metric](ground_truth, events, data=data)
            for metric in metrics
        }

        return pd.Series(scores)



import pytest

from orion.evaluation.common import _any_overlap, _overlap


@pytest.fixture
def expected():
    return {
        "true": [(1, 5), (1, 5), (1, 5)],
        "false": [(2, 4), (2, 4), (2, 4), (2, 2)]
    }


@pytest.fixture
def observed():
    return {
        "true": [(1, 4), (2, 5), (3, 4)],
        "false": [(1, 2), (4, 5), (5, 6), (2, 2)]
    }


def test__overlap_true(expected, observed):
    expected_true = expected["true"]
    observed_true = observed["true"]
    assert all(_overlap(ex, ob) for ex, ob in zip(expected_true, observed_true))


def test__overlap_false(expected, observed):
    expected_false = expected["false"]
    observed_false = observed["false"]
    assert not all(_overlap(ex, ob) for ex, ob in zip(expected_false, observed_false))


def test__any_overlap_true(expected, observed):
    part = expected["true"][0]
    interval = observed["true"]

    expected_return = 1
    returned = _any_overlap(part, interval)
    assert returned == expected_return


def test__any_overlap_false(expected, observed):
    part = expected["false"][0]
    interval = observed["false"]

    expected_return = 0
    returned = _any_overlap(part, interval)
    assert returned == expected_return



"""Orion Evaluation subpackage.

This subpackage contains all the code related to the
Orion Evaluation usage.
"""
from orion.evaluation.contextual import (
    contextual_accuracy, contextual_confusion_matrix, contextual_f1_score, contextual_precision,
    contextual_recall)
from orion.evaluation.point import (
    point_accuracy, point_confusion_matrix, point_f1_score, point_precision, point_recall)

CONTEXTUAL_METRICS = {
    'accuracy': contextual_accuracy,
    'f1': contextual_f1_score,
    'recall': contextual_recall,
    'precision': contextual_precision,
}

POINT_METRICS = {
    'accuracy': point_accuracy,
    'f1': point_f1_score,
    'recall': point_recall,
    'precision': point_precision,
}

__all__ = (
    'contextual_accuracy', 'contextual_confusion_matrix', 'contextual_f1_score',
    'contextual_precision', 'contextual_recall', 'CONTEXTUAL_METRICS',
    'point_accuracy', 'point_confusion_matrix', 'point_f1_score', 'point_precision',
    'point_recall', 'POINT_METRICS'
)



from orion.evaluation.common import _accuracy, _f1_score, _precision, _recall, _weighted_segment


def _point_partition(expected, observed, start=None, end=None):
    expected = set(expected)
    observed = set(observed)

    edge_start = min(expected.union(observed))
    if start is not None:
        edge_start = start

    edge_end = max(expected.union(observed))
    if end is not None:
        edge_end = end

    length = int(edge_end) - int(edge_start) + 1

    expected_parts = [0] * length
    observed_parts = [0] * length

    for edge in expected:
        expected_parts[edge - edge_start] = 1

    for edge in observed:
        observed_parts[edge - edge_start] = 1

    return expected_parts, observed_parts, None


def point_confusion_matrix(expected, observed, data=None, start=None, end=None):
    """Compute the confusion matrix between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of timestamps):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        observed (DataFrame or list of timestamps):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.

    Returns:
        tuple:
            number of true negative, false positive, false negative, true positive.
    """

    def _ws(x, y, z, w):
        return _weighted_segment(x, y, _point_partition, z, w)

    if data is not None:
        start = data['timestamp'].min()
        end = data['timestamp'].max()

    if not isinstance(expected, list):
        expected = list(expected['timestamp'])
    if not isinstance(observed, list):
        observed = list(observed['timestamp'])

    return _ws(expected, observed, start, end)


def point_accuracy(expected, observed, data=None, start=None, end=None):
    """Compute an accuracy score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of timestamps):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        observed (DataFrame or list of timestamps):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.

    Returns:
        float:
            Accuracy score between the ground truth and detected anomalies.
    """
    return _accuracy(expected, observed, data, start, end, cm=point_confusion_matrix)


def point_precision(expected, observed, data=None, start=None, end=None):
    """Compute an precision score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of timestamps):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        observed (DataFrame or list of timestamps):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.

    Returns:
        float:
            Precision score between the ground truth and detected anomalies.
    """
    return _precision(expected, observed, data, start, end, cm=point_confusion_matrix)


def point_recall(expected, observed, data=None, start=None, end=None):
    """Compute an recall score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of timestamps):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        observed (DataFrame or list of timestamps):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.

    Returns:
        float:
            Recall score between the ground truth and detected anomalies.
    """
    return _recall(expected, observed, data, start, end, cm=point_confusion_matrix)


def point_f1_score(expected, observed, data=None, start=None, end=None):
    """Compute an f1 score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of timestamps):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        observed (DataFrame or list of timestamps):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            one column: timestamp.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.

    Returns:
        float:
            F1 score between the ground truth and detected anomalies.
    """
    return _f1_score(expected, observed, data, start, end, cm=point_confusion_matrix)



# -*- coding: utf-8 -*-

import logging
import math
import tempfile
from typing import Optional, Union

import numpy as np
import pandas as pd
import tensorflow as tf
from mlstars.utils import import_object
from numpy import ndarray
from scipy import stats
from tensorflow.keras import Model

from orion.primitives.timeseries_errors import reconstruction_errors

LOGGER = logging.getLogger(__name__)
tf.keras.backend.set_floatx('float64')

LOSS_NAMES = [
    ['cx_loss', 'cx_real', 'cx_fake', 'cx_gp'],
    ['cz_loss', 'cz_real', 'cz_fake', 'cz_gp'],
    ['eg_loss', 'eg_cx_fake', 'eg_cz_fake', 'eg_mse']
]


def build_layer(layer: dict, hyperparameters: dict):
    layer_class = import_object(layer['class'])
    layer_kwargs = layer['parameters'].copy()
    # TODO: Upgrade to using tf.keras.layers.Wrapper in mlprimitives.
    if issubclass(layer_class, tf.keras.layers.Wrapper):
        layer_kwargs['layer'] = build_layer(layer_kwargs['layer'], hyperparameters)

    for key, value in layer_kwargs.items():
        if isinstance(value, str):
            layer_kwargs[key] = hyperparameters.get(value, value)

    return layer_class(**layer_kwargs)


class TadGAN:
    """TadGAN model for time series reconstruction.

    Args:
        layers_encoder (list):
            List containing layers of encoder.
        layers_generator (list):
            List containing layers of generator.
        layers_critic_x (list):
            List containing layers of critic_x.
        layers_critic_z (list):
            List containing layers of critic_z.
        input_shape (tuple):
            Optional. Tuple denoting the shape of an input sample.
        latent_dim (int):
            Optional. Integer denoting dimension of latent space. Default 20.
        target_shape (tuple):
            Optional. Tuple denoting the shape of an output sample.
        encoder_input_shape (tuple):
            Shape of encoder input.
        generator_input_shape (tuple):
            Shape of generator input.
        critic_x_input_shape (tuple):
            Shape of critic_x input.
        critic_z_input_shape (tuple):
            Shape of critic_z input.
        optimizer (str):
            String denoting the keras optimizer.
        learning_rate (float):
            Optional. Float denoting the learning rate of the optimizer. Default 0.005.
        epochs (int):
            Optional. Integer denoting the number of epochs. Default 50.
        latent_dim (int):
            Optional. Integer denoting dimension of latent space. Default 20.
        iterations_critic (int):
            Optional. Integer denoting the number of critic training steps per one
            Generator/Encoder training step. Default 5.
        shuffle (bool):
            Whether to shuffle the dataset for each epoch. Default True.
        verbose (int):
            Verbosity mode where 0 = silent, 1 = progress bar, 2 = one line per epoch. Default 0.
        detailed_losses (bool):
            Whether to output all loss values in verbose mode. Default False.
        **hyperparameters (dict):
            Optional. Dictionary containing any additional inputs.
    """
    @staticmethod
    def _build_model(hyperparameters: dict, layers: list,
                     input_shape: tuple, name: str) -> Model:
        x = tf.keras.layers.Input(shape=input_shape)
        model = tf.keras.Sequential(name=name)

        for layer in layers:
            built_layer = build_layer(layer, hyperparameters)
            model.add(built_layer)

        return Model(x, model(x))

    @staticmethod
    def _wasserstein_loss(y_true, y_pred):
        return tf.reduce_mean(y_true * y_pred)

    @staticmethod
    def _gradient_penalty_loss_wrapper(critic):
        def _gradient_penalty_loss(real, fake):
            # Random weighted average to create interpolated signals.
            batch_size = tf.shape(real)[0]
            alpha = tf.random.uniform([batch_size, 1, 1], dtype=tf.float64)
            interpolated = (alpha * real) + ((1 - alpha) * fake)

            # Get the critic output for this interpolated signal.
            with tf.GradientTape() as gp_tape:
                gp_tape.watch(interpolated)
                validity_interpolated = critic(interpolated)

            # Calculate the gradients w.r.t to this interpolated signal.
            grads = gp_tape.gradient(validity_interpolated, [interpolated])[0]

            # Calculate the norm of the gradients.
            norm = tf.sqrt(tf.reduce_sum(tf.square(grads), axis=np.arange(1, len(grads.shape))))
            gradient_penalty = tf.reduce_mean((norm - 1.0) ** 2)
            return gradient_penalty

        return _gradient_penalty_loss

    def __init__(self, layers_encoder: list, layers_generator: list, layers_critic_x: list,
                 layers_critic_z: list, optimizer: str, input_shape: Optional[tuple] = None,
                 target_shape: Optional[tuple] = None, latent_dim: int = 20,
                 learning_rate: float = 0.005, epochs: int = 50, batch_size: int = 64,
                 iterations_critic: int = 5, shuffle: bool = True, detailed_losses: bool = False,
                 verbose: Union[int, bool] = True, **hyperparameters):
        """Initialize the TadGAN model."""
        super(TadGAN, self).__init__()

        # Required model hyperparameters to construct model.
        self.layers_encoder, self.layers_generator = layers_encoder, layers_generator
        self.layers_critic_x, self.layers_critic_z = layers_critic_x, layers_critic_z

        # Optional model hyperparameters.
        self.shape = input_shape
        self.latent_dim = latent_dim
        self.target_shape = target_shape
        self.hyperparameters = hyperparameters

        # Model training hyperparameters.
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.epochs = epochs
        self.batch_size = batch_size
        self.iterations_critic = iterations_critic
        self.shuffle = shuffle
        self.verbose = verbose
        self.detailed_losses = detailed_losses
        self.fitted = False

    def __getstate__(self):
        networks = ['critic_x', 'critic_z', 'encoder', 'generator']
        modules = ['critic_x_model', 'critic_z_model', 'encoder_generator_model',
                   'critic_x_optimizer', 'critic_z_optimizer', 'encoder_generator_optimizer']

        state = self.__dict__.copy()

        for module in modules:
            del state[module]

        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                tf.keras.models.save_model(state.pop(network), fd.name, overwrite=True)
                state[network + '_str'] = fd.read()

        return state

    def __setstate__(self, state: dict):
        networks = ['critic_x', 'critic_z', 'encoder', 'generator']
        for network in networks:
            with tempfile.NamedTemporaryFile(suffix='.hdf5', delete=False) as fd:
                fd.write(state.pop(network + '_str'))
                fd.flush()

                state[network] = tf.keras.models.load_model(fd.name)

        self.__dict__ = state

    def _augment_hyperparameters(self, X: ndarray, y: ndarray, **kwargs) -> dict:
        shape = np.asarray(X)[0].shape
        length, input_dim = shape
        target_shape = np.asarray(y)[0].shape
        output_dim = target_shape[1]

        # Infers the shape.
        self.shape = self.shape or shape
        self.target_shape = self.target_shape or target_shape
        self.latent_shape = (self.latent_dim, output_dim)

        kwargs.update({
            'shape': self.shape,
            'target_shape': self.target_shape,
            'input_dim': input_dim,
            'output_dim': output_dim,
            'generator_reshape_dim': length // 2,
            'generator_reshape_shape': (length // 2, output_dim),
            'encoder_reshape_shape': self.latent_shape
        })
        return kwargs

    def _set_shapes(self) -> None:
        self.encoder_input_shape = self.shape
        self.generator_input_shape = self.latent_shape
        self.critic_x_input_shape = self.target_shape
        self.critic_z_input_shape = self.latent_shape

    def _format_losses(self, losses: list) -> dict:
        """Format losses into dictionary mapping metric names to their current value."""
        output = dict()
        if self.detailed_losses:
            for i in range(len(losses)):
                if not isinstance(losses[i], list):
                    losses[i] = [losses[i]]
                for j in range(len(losses[i])):
                    output[LOSS_NAMES[i][j]] = losses[i][j]
        else:
            for i in range(len(losses)):
                output[LOSS_NAMES[i][0]] = losses[i][0]

        return output

    def _build_tadgan(self, **kwargs) -> None:
        hyperparameters = self.hyperparameters.copy()
        hyperparameters.update(kwargs)

        # Models
        self.encoder = self._build_model(
            hyperparameters, self.layers_encoder, self.encoder_input_shape, name='encoder')
        self.generator = self._build_model(
            hyperparameters, self.layers_generator, self.generator_input_shape, name='generator')
        self.critic_x = self._build_model(
            hyperparameters, self.layers_critic_x, self.critic_x_input_shape, name='critic_x')
        self.critic_z = self._build_model(
            hyperparameters, self.layers_critic_z, self.critic_z_input_shape, name='critic_z')

        # Optimizers
        self.critic_x_optimizer = import_object(self.optimizer)(self.learning_rate)
        self.critic_z_optimizer = import_object(self.optimizer)(self.learning_rate)
        self.encoder_generator_optimizer = import_object(self.optimizer)(self.learning_rate)

        x = tf.keras.Input(shape=self.shape)
        y = tf.keras.Input(shape=self.target_shape)
        z = tf.keras.Input(shape=self.latent_shape)

        self.generator.trainable = False
        self.encoder.trainable = False

        # Critic x model
        x_ = self.generator(z)
        cx_real = self.critic_x(y)
        cx_fake = self.critic_x(x_)
        self.critic_x_model = Model(inputs=[y, z], outputs=[cx_real, cx_fake, x_])
        self.critic_x_model.compile(loss=[self._wasserstein_loss,
                                          self._wasserstein_loss,
                                          self._gradient_penalty_loss_wrapper(self.critic_x)
                                          ],
                                    optimizer=self.critic_x_optimizer,
                                    loss_weights=[1, 1, 10])

        # Critic z model
        z_ = self.encoder(x)
        cz_real = self.critic_z(z)
        cz_fake = self.critic_z(z_)
        self.critic_z_model = Model(inputs=[x, z], outputs=[cz_real, cz_fake, z_])
        self.critic_z_model.compile(loss=[self._wasserstein_loss,
                                          self._wasserstein_loss,
                                          self._gradient_penalty_loss_wrapper(self.critic_z)
                                          ],
                                    optimizer=self.critic_z_optimizer,
                                    loss_weights=[1, 1, 10])

        self.critic_x.trainable = False
        self.critic_z.trainable = False
        self.generator.trainable = True
        self.encoder.trainable = True

        x_ = self.generator(z)
        cx_fake = self.critic_x(x_)
        z_ = self.encoder(x)
        cz_fake = self.critic_z(z_)
        x_rec_ = self.generator(z_)

        self.encoder_generator_model = Model([x, z], [cx_fake, cz_fake, x_rec_])
        self.encoder_generator_model.compile(loss=[self._wasserstein_loss,
                                                   self._wasserstein_loss,
                                                   'mse'
                                                   ],
                                             optimizer=self.encoder_generator_optimizer,
                                             loss_weights=[1, 1, 10])

    def _fit(self, data: tuple):
        X_train, y_train = data
        minibatch_size = self.batch_size * self.iterations_critic
        num_minibatch = X_train.shape[0] // minibatch_size
        z_shape = (self.batch_size, self.latent_shape[0], self.latent_shape[1])

        # Wasserstein GAN formulation.
        fake = tf.ones((self.batch_size, 1), dtype=tf.float64)
        real = -tf.ones((self.batch_size, 1), dtype=tf.float64)

        indices = np.arange(X_train.shape[0])
        for epoch in range(1, self.epochs + 1):
            if self.shuffle:
                np.random.shuffle(indices)

            x_shuffled = X_train[indices]
            y_shuffled = y_train[indices]
            epoch_cx_loss, epoch_cz_loss, epoch_eg_loss = [], [], []

            for i in range(num_minibatch):
                x_minibatch = x_shuffled[i * minibatch_size: (i + 1) * minibatch_size]
                y_minibatch = y_shuffled[i * minibatch_size: (i + 1) * minibatch_size]

                for j in range(self.iterations_critic):
                    x = x_minibatch[j * self.batch_size: (j + 1) * self.batch_size]
                    y = y_minibatch[j * self.batch_size: (j + 1) * self.batch_size]
                    z = tf.random.normal(shape=z_shape, dtype=tf.float64)
                    epoch_cx_loss.append(
                        self.critic_x_model.train_on_batch([y, z], [real, fake, y]))
                    epoch_cz_loss.append(
                        self.critic_z_model.train_on_batch([x, z], [real, fake, z]))

                epoch_eg_loss.append(
                    self.encoder_generator_model.train_on_batch([x, z], [real, real, y]))

            epoch_cx_loss = np.round(np.mean(np.array(epoch_cx_loss), axis=0), 4)
            epoch_cz_loss = np.round(np.mean(np.array(epoch_cz_loss), axis=0), 4)
            epoch_eg_loss = np.round(np.mean(np.array(epoch_eg_loss), axis=0), 4)
            losses = self._format_losses([epoch_cx_loss, epoch_cz_loss, epoch_eg_loss])
            if self.verbose:
                print('Epoch: {}/{}, Losses: {}'.format(epoch, self.epochs, losses))

    def fit(self, X: ndarray, y: Optional[ndarray] = None, **kwargs) -> None:
        """Fit the TadGAN model.

        Args:
            X (ndarray):
                N-dimensional array containing the input training sequences for the model.
            y (ndarray):
                N-dimensional array containing the target sequences we want to reconstruct.
            **kwargs (dict):
                Optional. Additional inputs.
        """
        if y is None:
            y = X.copy()
        X, y = X.astype(np.float64), y.astype(np.float64)

        # Infer dimensions and compile model.
        if not self.fitted:
            kwargs = self._augment_hyperparameters(X, y, **kwargs)
            self._set_shapes()
            self._build_tadgan(**kwargs)

        self._fit((X, y))
        self.fitted = True

    def predict(self, X: ndarray, y: Optional[ndarray] = None) -> tuple:
        """Predict using TadGAN model.

        Args:
            X (ndarray):
                N-dimensional array containing the input training sequences for the model.
            y (ndarray):
                N-dimensional array containing the target sequences we want to reconstruct.
        Returns:
            ndarray:
                N-dimensional array containing the reconstructions for each input sequence.
            ndarray:
                N-dimensional array containing the critic scores for each input sequence.
        """
        if y is None:
            y = X.copy()
        X, y = X.astype(np.float64), y.astype(np.float64)

        z_ = self.encoder.predict(X)
        y_hat = self.generator.predict(z_)
        critic = self.critic_x.predict(y)

        return y_hat, critic


def _compute_critic_score(critics: ndarray, smooth_window: int) -> ndarray:
    """Compute an array of anomaly scores.

    Args:
        critics (ndarray):
            Critic values.
        smooth_window (int):
            Smooth window that will be applied to compute smooth errors.

    Returns:
        ndarray:
            Array of anomaly scores.
    """
    critics = np.asarray(critics)
    l_quantile = np.quantile(critics, 0.25)
    u_quantile = np.quantile(critics, 0.75)
    in_range = np.logical_and(critics >= l_quantile, critics <= u_quantile)
    critic_mean = np.mean(critics[in_range])
    critic_std = np.std(critics)

    z_scores = np.absolute((np.asarray(critics) - critic_mean) / critic_std) + 1
    z_scores = pd.Series(z_scores).rolling(
        smooth_window, center=True, min_periods=smooth_window // 2).mean().values

    return z_scores


def score_anomalies(y: ndarray, y_hat: ndarray, critic: ndarray, index: ndarray,
                    score_window: int = 10, critic_smooth_window: int = None,
                    error_smooth_window: int = None, smooth: bool = True,
                    rec_error_type: str = "point", comb: str = "mult",
                    lambda_rec: float = 0.5) -> tuple:
    """Compute an array of anomaly scores.

    Anomaly scores are calculated using a combination of reconstruction error and critic score.

    Args:
        y (ndarray):
            Ground truth.
        y_hat (ndarray):
            Predicted values. Each timestamp has multiple predictions.
        index (ndarray):
            time index for each y (start position of the window)
        critic (ndarray):
            Critic score. Each timestamp has multiple critic scores.
        score_window (int):
            Optional. Size of the window over which the scores are calculated.
            If not given, 10 is used.
        critic_smooth_window (int):
            Optional. Size of window over which smoothing is applied to critic.
            If not given, 200 is used.
        error_smooth_window (int):
            Optional. Size of window over which smoothing is applied to error.
            If not given, 200 is used.
        smooth (bool):
            Optional. Indicates whether errors should be smoothed.
            If not given, `True` is used.
        rec_error_type (str):
            Optional. The method to compute reconstruction error. Can be one of
            `["point", "area", "dtw"]`. If not given, 'point' is used.
        comb (str):
            Optional. How to combine critic and reconstruction error. Can be one
            of `["mult", "sum", "rec"]`. If not given, 'mult' is used.
        lambda_rec (float):
            Optional. Used if `comb="sum"` as a lambda weighted sum to combine
            scores. If not given, 0.5 is used.

    Returns:
        ndarray:
            Array of anomaly scores.
    """

    critic_smooth_window = critic_smooth_window or math.trunc(y.shape[0] * 0.01)
    error_smooth_window = error_smooth_window or math.trunc(y.shape[0] * 0.01)

    step_size = 1  # expected to be 1

    true_index = index  # no offset

    true = [item[0] for item in y.reshape((y.shape[0], -1))]

    for item in y[-1][1:]:
        true.extend(item)

    critic_extended = list()
    for c in critic:
        critic_extended.extend(np.repeat(c, y_hat.shape[1]).tolist())

    critic_extended = np.asarray(critic_extended).reshape((-1, y_hat.shape[1]))

    critic_kde_max = []
    pred_length = y_hat.shape[1]
    num_errors = y_hat.shape[1] + step_size * (y_hat.shape[0] - 1)

    for i in range(num_errors):
        critic_intermediate = []

        for j in range(max(0, i - num_errors + pred_length), min(i + 1, pred_length)):
            critic_intermediate.append(critic_extended[i - j, j])

        if len(critic_intermediate) > 1:
            discr_intermediate = np.asarray(critic_intermediate)
            try:
                critic_kde_max.append(discr_intermediate[np.argmax(
                    stats.gaussian_kde(discr_intermediate)(critic_intermediate))])
            except np.linalg.LinAlgError:
                critic_kde_max.append(np.median(discr_intermediate))
        else:
            critic_kde_max.append(np.median(np.asarray(critic_intermediate)))

    # Compute critic scores
    critic_scores = _compute_critic_score(critic_kde_max, critic_smooth_window)

    # Compute reconstruction scores
    rec_scores, predictions = reconstruction_errors(
        y, y_hat, step_size, score_window, error_smooth_window, smooth, rec_error_type)

    rec_scores = stats.zscore(rec_scores)
    rec_scores = np.clip(rec_scores, a_min=0, a_max=None) + 1

    # Combine the two scores
    if comb == "mult":
        final_scores = np.multiply(critic_scores, rec_scores)

    elif comb == "sum":
        final_scores = (1 - lambda_rec) * (critic_scores - 1) + lambda_rec * (rec_scores - 1)

    elif comb == "rec":
        final_scores = rec_scores

    else:
        raise ValueError(
            'Unknown combination specified {}, use "mult", "sum", or "rec" instead.'.format(comb))

    true = [[t] for t in true]
    return final_scores, true_index, true, predictions



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

try:
    with open('README.md', encoding='utf-8') as readme_file:
        readme = readme_file.read()
except IOError:
    readme = ''

try:
    with open('HISTORY.md', encoding='utf-8') as history_file:
        history = history_file.read()
except IOError:
    history = ''


install_requires = [
    "tensorflow>=2.0,<2.5;python_version<'3.8'",
    "tensorflow>=2.2,<2.5;python_version>='3.8'",
    's3fs>=0.2.2,<0.5',
    'mlblocks>=0.4.1,<0.5',
    'ml-stars>=0.1.1,<0.2',
    'numpy>=1.17.4,<1.19',
    'pandas>=1,<2',
    'scikit-learn>=0.22,<1.2',
    'tabulate>=0.8.3,<0.9',
    'numba>=0.48,<0.52',
    'pyts>=0.9,<0.11',
    'azure-cognitiveservices-anomalydetector>=0.3,<0.4',
    'xlsxwriter>=1.3.6,<1.4',
    'tqdm>=4.36.1',

    # fix conflict
    'h5py<2.11.0,>=2.10.0',
    'protobuf<4',
    
    # fails on python 3.6
    'opencv-python<4.7',
]

setup_requires = [
    'pytest-runner>=2.11.1',
]

tests_require = [
    'pytest>=3.4.2',
    'pytest-cov>=2.6.0',
    'rundoc>=0.4.3,<0.5',
]

development_requires = [
    # general
    'pip>=9.0.1',
    'bumpversion>=0.5.3,<0.6',
    'watchdog>=0.8.3,<0.11',

    # docs
    'docutils>=0.12,<0.18',
    'm2r2>=0.2.5,<0.3',
    'nbsphinx>=0.5.0,<0.7',
    'Sphinx>=3,<3.3',
    'pydata-sphinx-theme<0.5',
    'autodocsumm>=0.1.10,<1',
    'markupsafe<2.1.0',
    'ipython>=6.5,<7.5',
    'Jinja2>=2,<3',

    # style check
    'flake8>=3.7.7,<4',
    'isort>=4.3.4,<5',

    # fix style issues
    'autoflake>=1.2,<2',
    'autopep8>=1.4.3,<2',
    'importlib-metadata<5',

    # distribute on PyPI
    'twine>=1.10.0,<4',
    'wheel>=0.30.0',

    # Advanced testing
    'coverage>=4.5.1,<6',
    'tox>=2.9.1,<4',
    'invoke',
]

setup(
    author="MIT Data To AI Lab",
    author_email='dailabmit@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Orion is a machine learning library built for unsupervised time series anomaly detection.",
    entry_points={
        'console_scripts': [
            'orion=orion.__main__:main'
        ],
        'mlblocks': [
            'primitives=orion:MLBLOCKS_PRIMITIVES',
            'pipelines=orion:MLBLOCKS_PIPELINES'
        ],
    },
    extras_require={
        'test': tests_require,
        'dev': development_requires + tests_require,
    },
    include_package_data=True,
    install_requires=install_requires,
    keywords='orion',
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    name='orion-ml',
    packages=find_packages(include=['orion', 'orion.*']),
    python_requires='>=3.6,<3.9',
    setup_requires=setup_requires,
    test_suite='tests',
    tests_require=tests_require,
    url='https://github.com/sintel-dev/Orion',
    version='0.5.0',
    zip_safe=False,
)



from orion.evaluation.common import (
    _accuracy, _any_overlap, _f1_score, _overlap, _precision, _recall, _weighted_segment)


def _overlap_segment(expected, observed, start=None, end=None):
    tp, fp, fn = 0, 0, 0

    observed_copy = observed.copy()

    for expected_seq in expected:
        found = False
        for observed_seq in observed:
            if _overlap(expected_seq, observed_seq):
                if not found:
                    tp += 1
                    found = True
                if observed_seq in observed_copy:
                    observed_copy.remove(observed_seq)

        if not found:
            fn += 1

    fp += len(observed_copy)

    return None, fp, fn, tp


def _contextual_partition(expected, observed, start=None, end=None):
    edges = set()

    if start is not None:
        edges.add(start)

    if end is not None:
        edges.add(end)

    for edge in expected + observed:
        edges.update(edge)

    partitions = list()
    edges = sorted(edges)
    last = edges[0]
    for edge in edges[1:]:
        partitions.append((last, edge))
        last = edge

    expected_parts = list()
    observed_parts = list()
    weights = list()
    for part in partitions:
        weights.append(part[1] - part[0])
        expected_parts.append(_any_overlap(part, expected))
        observed_parts.append(_any_overlap(part, observed))

    return expected_parts, observed_parts, weights


def _pad(lst):
    return [(part[0], part[1] + 1) for part in lst]


def contextual_confusion_matrix(expected, observed, data=None,
                                start=None, end=None, weighted=True):
    """Compute the confusion matrix between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of tuples):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        observed (DataFrame or list of tuples):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.
        weighted (boolean):
            Flag to represent which algorithm to use.
            If true use weighted segment algorithm, else use overlap segment.

    Returns:
        tuple:
            number of true negative, false positive, false negative, true positive.
    """

    def _ws(x, y, z, w):
        return _weighted_segment(x, y, _contextual_partition, z, w)

    if weighted:
        function = _ws
    else:
        function = _overlap_segment

    if data is not None:
        start = data['timestamp'].min()
        end = data['timestamp'].max()

    if not isinstance(expected, list):
        expected = list(expected[['start', 'end']].itertuples(index=False))
    if not isinstance(observed, list):
        observed = list(observed[['start', 'end']].itertuples(index=False))

    expected = _pad(expected)
    observed = _pad(observed)

    return function(expected, observed, start, end)


def contextual_accuracy(expected, observed, data=None, start=None, end=None, weighted=True):
    """Compute an accuracy score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of tuples):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        observed (DataFrame or list of tuples):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.
        weighted (boolean):
            Flag to represent which algorithm to use.
            If true use weighted segment algorithm, else use overlap segment.

    Returns:
        float:
            Accuracy score between the ground truth and detected anomalies.
    """
    def _cm(x, y, z, w, f):
        return contextual_confusion_matrix(x, y, z, w, f, weighted)
    return _accuracy(expected, observed, data, start, end, _cm)


def contextual_precision(expected, observed, data=None, start=None, end=None, weighted=True):
    """Compute an precision score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of tuples):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        observed (DataFrame or list of tuples):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.
        weighted (boolean):
            Flag to represent which algorithm to use.
            If true use weighted segment algorithm, else use overlap segment.

    Returns:
        float:
            Precision score between the ground truth and detected anomalies.
    """
    def _cm(x, y, z, w, f):
        return contextual_confusion_matrix(x, y, z, w, f, weighted)
    return _precision(expected, observed, data, start, end, _cm)


def contextual_recall(expected, observed, data=None, start=None, end=None, weighted=True):
    """Compute an recall score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of tuples):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        observed (DataFrame or list of tuples):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.
        weighted (boolean):
            Flag to represent which algorithm to use.
            If true use weighted segment algorithm, else use overlap segment.

    Returns:
        float:
            Recall score between the ground truth and detected anomalies.
    """
    def _cm(x, y, z, w, f):
        return contextual_confusion_matrix(x, y, z, w, f, weighted)
    return _recall(expected, observed, data, start, end, _cm)


def contextual_f1_score(expected, observed, data=None, start=None, end=None, weighted=True):
    """Compute an f1 score between the ground truth and the detected anomalies.

    Args:
        expected (DataFrame or list of tuples):
            Ground truth passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        observed (DataFrame or list of tuples):
            Detected anomalies passed as a ``pandas.DataFrame`` or list containing
            two columns: start and stop.
        data (DataFrame):
            Original data, passed as a ``pandas.DataFrame`` containing timestamp.
            Used to extract start and end.
        start (int):
            Minimum timestamp of the original data.
        end (int):
            Maximum timestamp of the original data.
        weighted (boolean):
            Flag to represent which algorithm to use.
            If true use weighted segment algorithm, else use overlap segment.

    Returns:
        float:
            F1 score between the ground truth and detected anomalies.
    """
    def _cm(x, y, z, w, f):
        return contextual_confusion_matrix(x, y, z, w, f, weighted)
    return _f1_score(expected, observed, data, start, end, _cm)
